<?php

declare(strict_types=1);

namespace Drupal\dwarl_mailer;

use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Utility\Error;

/**
 * Email manager for sending emails.
 */
final class EmailManager implements EmailManagerInterface {
  use StringTranslationTrait;

  /**
   * Constructs an EmailManager object.
   */
  public function __construct(
    private readonly RendererInterface $renderer,
    private readonly LanguageManagerInterface $languageManager,
    private readonly MailManagerInterface $mailManager,
    private readonly LoggerChannelInterface $logger,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function sendCodeEmail(string $email, string $code, string $text): void {

    $siteMail = \Drupal::config('system.site')->get('mail');

    $params = [
      'to' => $email,
      'from' => $siteMail,
      'subject' => (string) $this->t('Your secret code is'),
      'body' => $this->codeEmailBody($code, $text),
      'plaintext' => $this->plainText($code, $text),
    ];

    $langCode = $this->languageManager->getDefaultLanguage()->getId();
    $result = $this->mailManager->mail(
      'dwarl_mailer',
      'dwarl_mailer_new_code',
      $email,
      $langCode,
      $params,
      $siteMail
    );

    $context = ['@email' => $email];
    if (isset($result['result']) && $result['result']) {
      $this->logger->notice('Send code email to @email', $context);
    }
    else {
      $this->logger->error('Failed to send code email to @email', $context);
      throw new \RuntimeException('Could not send email');
    }

  }

  /**
   * Build code email body.
   *
   * @param string $code
   *   The code.
   * @param string $text
   *   The text.
   *
   * @return string
   *   The body.
   *
   * @throws \RuntimeException
   *   If fails to render.
   */
  private function codeEmailBody(string $code, string $text): string {
    $body = [
      '#theme' => 'dwarl_mailer_code_email',
      '#text' => $text,
      '#code_start' => substr($code, 0, 3),
      '#code_end' => substr($code, 3, 3),
    ];

    try {
      return (string) $this->renderer->render($body);
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception);
      throw new \RuntimeException('Mailer error');
    }

  }

  /**
   * Get plain text email.
   *
   * @param string $code
   *   The code.
   * @param string $text
   *   The text.
   *
   * @return string
   *   The plain text email.
   */
  private function plainText(string $code, string $text): string {
    return $text . $code . PHP_EOL . PHP_EOL;
  }

  /**
   * {@inheritDoc}
   */
  public static function hookMail(string $key, array &$message, array $params): void {
    $message['module'] = 'dwarl_mailer';
    $message['key'] = $key;
    $message['subject'] = $params['subject'];
    $message['plaintext'] = $params['plaintext'];

    $mailLineEndings = Settings::get('mail_line_endings', PHP_EOL);
    $separator = $mailLineEndings . $mailLineEndings;

    $message['body'] = explode($separator, $params['body']);

    if ($replyTo = $params['reply_to'] ?? NULL) {
      $message['headers']['Reply-to'] = $replyTo;
    }
    if ($cc = $params['cc'] ?? NULL) {
      $message['headers']['cc'] = $cc;
    }
    if ($bcc = $params['bcc'] ?? NULL) {
      $message['headers']['bcc'] = $bcc;
    }
  }

}
