<?php

declare(strict_types=1);

namespace Drupal\dwarl\Access;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Routing\Access\AccessInterface;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\Exception\DwarlStoreException;
use Drupal\dwarl\Exception\StoreExpiredException;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Base access check class.
 */
abstract class BaseAccessCheck implements AccessInterface {

  /**
   * BaseAccessCheck constructor.
   */
  public function __construct(
    private readonly AnonymousStoreInterface $anonymousStore,
    private readonly LoggerInterface $logger,
  ) {
  }

  /**
   * Check access.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   The access result.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function access(Request $request): AccessResultInterface {

    // Deny access if no session id.
    $sessionId = $request->headers->get('session-id');
    if (empty($sessionId)) {
      return AccessResult::forbidden('No session id');
    }

    $this->logger->info('Got session id: ' . $sessionId);

    try {
      $this->anonymousStore->init($sessionId);
    }
    catch (StoreExpiredException) {
      return AccessResult::forbidden('Session expired');
    }
    catch (DwarlStoreException $exception) {
      Error::logException($this->logger, $exception);
      return AccessResult::forbidden('Store error');
    }

    try {
      $this->anonymousStore->getUser();
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception);
      return AccessResult::forbidden('No user');
    }

    if ($this->anonymousStore->getRealm() !== $this->getRealm()) {
      return AccessResult::forbidden('Bad realm');
    }

    // Grant access.
    return AccessResult::allowed();
  }

  /**
   * Get the realm of this access checker.
   *
   * Stop users from being able to use a session from a forgot password
   * flow in the register flow, for example.
   *
   * @return string
   *   The name of this realm.
   */
  abstract protected function getRealm(): string;

}
