<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\Exception\MissingParameterException;
use Drupal\dwarl\Exception\StringTooShortException;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Dwarl base controller.
 */
abstract class DwarlControllerBase extends ControllerBase {

  /**
   * DwarlControllerBase constructor.
   */
  public function __construct(
    private readonly Request $request,
    private readonly LoggerInterface $logger,
  ) {
  }

  /**
   * Get value from request.
   *
   * @param string $key
   *   The key.
   *
   * @return mixed|null
   *   The value.
   */
  protected function getFromRequest(string $key): mixed {
    $value = $this->request->get($key);
    if ($value) {
      return $value;
    }
    $request = Json::decode(file_get_contents("php://input"));
    return $request[$key] ?? NULL;
  }

  /**
   * Get required parameter.
   *
   * @param string $key
   *   The key.
   *
   * @return mixed
   *   The value.
   *
   * @throws \Drupal\dwarl\Exception\MissingParameterException
   */
  protected function getRequiredParameter(string $key): mixed {
    $value = $this->getFromRequest($key);
    if ($value === NULL) {
      throw new MissingParameterException($key);
    }
    return $value;
  }

  /**
   * Get required string.
   *
   * @param string $key
   *   The key.
   * @param int|null $minLength
   *   The minimum length of string required.
   *
   * @return string
   *   The string value.
   *
   * @throws \Drupal\dwarl\Exception\MissingParameterException
   */
  protected function getRequiredString(string $key, ?int $minLength = NULL): string {
    $string = (string) $this->getRequiredParameter($key);
    if ($minLength && strlen($string) < $minLength) {
      throw new StringTooShortException($key, $minLength);
    }
    return $string;
  }

  /**
   * Get password from request.
   *
   * @return string
   *   The password.
   */
  protected function getPasswordFromRequest(): string {
    return $this->getRequiredString('password', PasswordController::PASSWORD_MIN_LENGTH);
  }

  /**
   * Return handled exception response.
   *
   * @param \Symfony\Component\HttpKernel\Exception\HttpException $exception
   *   The exception.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   */
  protected function handledException(HttpException $exception): JsonResponse {
    $response = ['error' => $exception->getMessage()];
    // @todo Make some of these loggable?
    return new JsonResponse($response, $exception->getStatusCode());
  }

  /**
   * Return unhandled exception response.
   *
   * @param \Throwable $exception
   *   The exception.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   */
  protected function unhandledException(\Throwable $exception): JsonResponse {
    Error::logException($this->logger, $exception);
    // @todo switch this back.
    // $response = ['error' => 'Internal Server Error'];
    $response = [
      'error' => $exception->getMessage(),
      'file' => $exception->getFile(),
      'line' => $exception->getLine(),
      'trace' => $exception->getTraceAsString(),
    ];
    return new JsonResponse($response, Response::HTTP_INTERNAL_SERVER_ERROR);
  }

  /**
   * Get json response with session id.
   *
   * @param string $sessionId
   *   The session id.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   */
  protected function sessionIdResponse(string $sessionId): JsonResponse {
    $response = new JsonResponse();
    $response->headers->set('session-id', $sessionId);
    $response->headers->set('Access-Control-Expose-Headers', ['session-id']);
    return $response;
  }

}
