<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\dwarl\Access\ForgotPasswordAccessCheck;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl\WebAuthn\AuthenticateRequestInterface;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Dwarl password controller.
 */
final class ForgotPasswordController extends DwarlControllerBase {

  /**
   * DwarlUserController constructor.
   */
  public function __construct(
    LoggerInterface $logger,
    private readonly Request $request,
    private readonly AnonymousStoreInterface $anonymousStore,
    private readonly JwtAuth $jwtAuth,
    private readonly RequestOptionsInterface $requestOptions,
    private readonly AuthenticateRequestInterface $authenticateRequest,
    private readonly OtpServiceInterface $otpService,
    private readonly UserService $userService,
    private readonly PasskeyStorageInterface $passkeyStorage,
  ) {
    parent::__construct($this->request, $logger);
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = $container->get('logger.channel.dwarl');

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $anonymousStore */
    $anonymousStore = $container->get('dwarl.anonymous_store');

    /** @var \Drupal\jwt\Authentication\Provider\JwtAuth $jwtAuth */
    $jwtAuth = $container->get('jwt.authentication.jwt');

    /** @var \Drupal\dwarl\WebAuthn\RequestOptionsInterface $requestOptions */
    $requestOptions = $container->get('dwarl.request_options');

    /** @var \Drupal\dwarl\WebAuthn\AuthenticateRequestInterface $authenticateRequest */
    $authenticateRequest = $container->get('dwarl.authenticate_request');

    /** @var \Drupal\dwarl\Otp\OtpServiceInterface $otpService */
    $otpService = $container->get('dwarl.otp_service');

    /** @var \Drupal\dwarl\UserService $userService */
    $userService = $container->get('dwarl.user_service');

    /** @var \Drupal\dwarl\PasskeyStorageInterface $passkeyStorage */
    $passkeyStorage = $container->get('dwarl.passkey_storage');

    return new self(
      $logger,
      $requestStack->getCurrentRequest(),
      $anonymousStore,
      $jwtAuth,
      $requestOptions,
      $authenticateRequest,
      $otpService,
      $userService,
      $passkeyStorage,
    );
  }

  /**
   * Forgot password.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   */
  public function forgotPassword(): JsonResponse {
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadByUserHandle($userHandle);
    $this->anonymousStore->init();

    if ($user && $user->isBlocked() === FALSE) {
      $this->anonymousStore->setUser($user);
      $this->anonymousStore->setRealm(ForgotPasswordAccessCheck::REALM);
      $this->otpService->generateOtp($this->anonymousStore, $user, OtpSendReason::ForgotPassword);
    }
    $data = ['status' => 'OTP sent if contact method exists.'];
    return $this->anonymousStore->getSessionIdResponse($data);
  }

  /**
   * Forgot password set password.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\UserSaveException
   */
  public function forgotPasswordSetPassword(): JsonResponse {
    $this->otpService->validateOtp($this->anonymousStore);
    $user = $this->anonymousStore->getUser();
    $password = $this->getPasswordFromRequest();
    $passkeys = $this->passkeyStorage->getUserPasskeys((int) $user->id());

    if ($passkeys) {
      // User has passkey so request verification before setting password.
      return new JsonResponse([
        'requestOptions' => $this->requestOptions->getRequestOptions($user->getEmail()),
      ]);
    }

    // Just set the password.
    $user->setPassword($password);
    $this->userService->saveUser($user);

    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();
    return new JsonResponse(['token' => $token]);
  }

  /**
   * Forgot password set password confirm.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  public function forgotPasswordSetPasswordConfirm(): JsonResponse {
    $password = $this->getRequiredString('password');
    $authenticationRequest = $this->getRequiredString('authentication_request');
    $user = $this->userService->currentUser();
    $userHandle = $user->getEmail();
    $this->authenticateRequest->execute($userHandle, $authenticationRequest);

    $user->setPassword($password);
    $user->save();

    $token = $this->jwtAuth->generateToken();
    return new JsonResponse(['token' => $token]);
  }

}
