<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\dwarl\Access\LostPasskeyAccessCheck;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\UserService;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Handle WebAuthn requests.
 */
final class LostPasskeyController extends DwarlControllerBase {

  /**
   * The controller constructor.
   */
  public function __construct(
    private readonly JwtAuth $jwtAuth,
    private readonly AnonymousStoreInterface $anonymousStore,
    Request $request,
    LoggerInterface $logger,
    private readonly OtpServiceInterface $otpService,
    private readonly UserService $userService,
    private readonly PasskeyStorageInterface $passkeyStorage,
  ) {
    parent::__construct($request, $logger);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    /** @var \Drupal\jwt\Authentication\Provider\JwtAuth $jwtAuth */
    $jwtAuth = $container->get('jwt.authentication.jwt');

    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $anonymousStore */
    $anonymousStore = $container->get('dwarl.anonymous_store');

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = \Drupal::service('logger.channel.dwarl');

    /** @var \Drupal\dwarl\Otp\OtpServiceInterface $otpService */
    $otpService = $container->get('dwarl.otp_service');

    /** @var \Drupal\dwarl\UserService $userService */
    $userService = $container->get('dwarl.user_service');

    /** @var \Drupal\dwarl\PasskeyStorageInterface $passkeyStorage */
    $passkeyStorage = $container->get('dwarl.passkey_storage');

    return new self(
      $jwtAuth,
      $anonymousStore,
      $requestStack->getCurrentRequest(),
      $logger,
      $otpService,
      $userService,
      $passkeyStorage,
    );
  }

  /**
   * Lost passkey.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Response with session id.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   */
  public function lostPasskey(): JsonResponse {
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadByUserHandle($userHandle);
    $this->anonymousStore->init();

    if ($user && $user->isBlocked() === FALSE) {
      $this->anonymousStore->setUser($user);
      $this->anonymousStore->setRealm(LostPasskeyAccessCheck::REALM);
      $this->otpService->generateOtp($this->anonymousStore, $user, OtpSendReason::LostPasskey);
    }

    $data = ['status' => 'OTP sent if contact method exists.'];
    return $this->anonymousStore->getSessionIdResponse($data);
  }

  /**
   * Lost passkey submit OTP.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function lostPasskeySubmitOtp(): JsonResponse {
    $this->otpService->validateOtp($this->anonymousStore);
    $user = $this->anonymousStore->getUser();
    $passkey = $this->passkeyStorage->getUserSinglePasskey((int) $user->id());
    $password = $user->getPassword();

    if ($password) {
      $this->anonymousStore->setStatus('password-required');
      return new JsonResponse(['status' => 'supply password', 'token' => NULL]);
    }

    // Delete this passkey as it's now lost.
    $this->passkeyStorage->deletePasskey($passkey);

    // Clear store now lost passkey flow is complete.
    $this->anonymousStore->expire();

    // Return a JWT token to allow the user to set a new auth method.
    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();
    return new JsonResponse(['status' => 'ok', 'token' => $token]);
  }

  /**
   * Lost passkey submit password.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function lostPasskeySubmitPassword(): JsonResponse {
    // Check access has been granted to this endpoint via the store status.
    $this->anonymousStore->validateStatus('password-required');
    $password = $this->getPasswordFromRequest();
    $user = $this->anonymousStore->getUser();
    $this->userService->validateUserPassword($user, $password);

    // Delete this passkey as it's now lost.
    $passkey = $this->passkeyStorage->getUserSinglePasskey((int) $user->id());
    $passkey->delete();

    // Clear store now lost passkey flow is complete.
    $this->anonymousStore->expire();

    // Generate a JWT token.
    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();
    return new JsonResponse(['token' => $token]);
  }

}
