<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl\WebAuthn\AuthenticateRequestInterface;
use Drupal\dwarl\WebAuthn\RegisterDeviceInterface;
use Drupal\dwarl\WebAuthn\RegisterDeviceOptionsInterface;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Drupal\public_key_credential_source\PublicKeyCredentialSourceInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Handle WebAuthn requests.
 */
final class PasskeyController extends DwarlControllerBase {

  /**
   * The controller constructor.
   */
  public function __construct(
    private readonly AuthenticateRequestInterface $authenticateRequest,
    private readonly RegisterDeviceOptionsInterface $registerDeviceOptions,
    private readonly RequestOptionsInterface $requestOptions,
    private readonly RegisterDeviceInterface $registerDevice,
    private readonly JwtAuth $jwtAuth,
    private readonly AnonymousStoreInterface $anonymousStore,
    Request $request,
    private readonly LoggerInterface $logger,
    private readonly UserService $userService,
  ) {
    parent::__construct($request, $logger);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = $container->get('logger.channel.dwarl');

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $anonymousStore */
    $anonymousStore = $container->get('dwarl.anonymous_store');

    /** @var \Drupal\jwt\Authentication\Provider\JwtAuth $jwtAuth */
    $jwtAuth = $container->get('jwt.authentication.jwt');

    /** @var \Drupal\dwarl\WebAuthn\AuthenticateRequestInterface $authenticateRequest */
    $authenticateRequest = $container->get('dwarl.authenticate_request');

    /** @var \Drupal\dwarl\WebAuthn\RegisterDeviceOptionsInterface $registerDeviceOptions */
    $registerDeviceOptions = $container->get('dwarl.register_device_options');

    /** @var \Drupal\dwarl\WebAuthn\RequestOptionsInterface $requestOptions */
    $requestOptions = $container->get('dwarl.request_options');

    /** @var \Drupal\dwarl\WebAuthn\RegisterDeviceInterface $registerDevice */
    $registerDevice = $container->get('dwarl.register_device');

    /** @var \Drupal\dwarl\UserService $userService */
    $userService = $container->get('dwarl.user_service');

    return new self(
      $authenticateRequest,
      $registerDeviceOptions,
      $requestOptions,
      $registerDevice,
      $jwtAuth,
      $anonymousStore,
      $requestStack->getCurrentRequest(),
      $logger,
      $userService,
    );
  }

  /**
   * Register device options.
   */
  public function registerDeviceOptions(): JsonResponse {
    return new JsonResponse([
      'data' => $this->registerDeviceOptions->getRegistrationOptions(),
    ]);
  }

  /**
   * Register device during registration.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\UserSaveException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  public function registrationRegisterDevice(): JsonResponse {
    $attestation = $this->getRequiredString('attestation');
    $user = $this->registerDevice->register($attestation);
    $token = $this->jwtAuth->generateToken();

    $args = ['@userId' => $user->id()];
    $this->logger->info('Register device for user @userId', $args);

    // Clear store now registration is complete.
    $this->anonymousStore->expire();

    return new JsonResponse(['token' => $token]);
  }

  /**
   * Register device for authenticated user.
   *
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   */
  public function authenticatedRegisterDevice(): JsonResponse {
    $attestation = $this->getRequiredString('attestation');
    $user = $this->registerDevice->addPasskey($attestation);
    $args = ['@userId' => $user->id()];
    $this->logger->info('Register device for user @userId', $args);
    return new JsonResponse();
  }

  /**
   * Request options.
   *
   * First login api call.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function requestOptions(): JsonResponse {
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadRequiredByUserHandle($userHandle);
    $this->anonymousStore->init()->setUser($user);
    $this->anonymousStore->setRealm('login');
    $data = ['data' => $this->requestOptions->getRequestOptions($userHandle)];
    return $this->anonymousStore->getSessionIdResponse($data);
  }

  /**
   * Authenticate request.
   *
   * Second login api call.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  public function authenticateRequest(): JsonResponse {
    $authenticationRequest = $this->getRequiredString('authentication_request');
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadRequiredByUserHandle($userHandle);
    $this->authenticateRequest->execute($user->getEmail(), $authenticationRequest);
    $token = $this->jwtAuth->generateToken();
    return new JsonResponse(['token' => $token]);
  }

  /**
   * Request set password options.
   *
   * @throws \Drupal\dwarl\Exception\MissingParameterException
   * @throws \Drupal\dwarl\Exception\InvalidPasswordException
   * @throws \JsonException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnException
   */
  public function requestSetPasswordOptions() {
    $currentPassword = $this->getRequiredString('current_password');
    $user = $this->userService->currentUser();
    $this->userService->validateUserPassword($user, $currentPassword);
    $userHandle = $user->getEmail();
    $options = $this->requestOptions->getRequestOptions($userHandle);
    return new JsonResponse(['data' => $options]);
  }

  /**
   * Set password request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  public function setPasswordRequest(): JsonResponse {
    $authenticationRequest = $this->getRequiredString('authentication_request');
    $password = $this->getPasswordFromRequest();
    $user = $this->userService->currentUser();
    $userHandle = $user->getEmail();
    $user = $this->authenticateRequest->execute($userHandle, $authenticationRequest);
    $user->setPassword($password);
    $user->save();
    $args = ['@userId' => $user->id()];
    $this->logger->info('Set password for user @userId', $args);
    return new JsonResponse();
  }

  /**
   * Delete passkey.
   *
   * @param \Drupal\public_key_credential_source\PublicKeyCredentialSourceInterface $public_key_credential_source
   *   The passkey, aka public key credential source.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   An empty 200 response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function deletePasskey(PublicKeyCredentialSourceInterface $public_key_credential_source): JsonResponse {
    $public_key_credential_source->delete();
    return new JsonResponse();
  }

}
