<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\dwarl\Exception\UserHasPasswordException;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Dwarl password controller.
 */
final class PasswordController extends DwarlControllerBase {

  public const PASSWORD_MIN_LENGTH = 6;

  /**
   * Dwarl private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  private PrivateTempStore $tempStore;

  /**
   * DwarlUserController constructor.
   */
  public function __construct(
    private readonly Request $request,
    private readonly AnonymousStoreInterface $anonymousStore,
    private readonly LoggerInterface $logger,
    private readonly JwtAuth $jwtAuth,
    PrivateTempStoreFactory $privateTempStoreFactory,
    private readonly RequestOptionsInterface $requestOptions,
    private readonly OtpServiceInterface $otpService,
    private readonly UserService $userService,
    private readonly PasskeyStorageInterface $passkeyStorage,
  ) {
    $this->tempStore = $privateTempStoreFactory->get('dwarl');
    parent::__construct($this->request, $logger);
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $anonymousStore */
    $anonymousStore = $container->get('dwarl.anonymous_store');

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = $container->get('logger.channel.dwarl');

    /** @var \Drupal\Core\TempStore\PrivateTempStoreFactory $privateTempStoreFactory */
    $privateTempStoreFactory = $container->get('tempstore.private');

    /** @var \Drupal\jwt\Authentication\Provider\JwtAuth $jwtAuth */
    $jwtAuth = $container->get('jwt.authentication.jwt');

    /** @var \Drupal\dwarl\WebAuthn\RequestOptionsInterface $requestOptions */
    $requestOptions = $container->get('dwarl.request_options');

    /** @var \Drupal\dwarl\Otp\OtpServiceInterface $otpService */
    $otpService = $container->get('dwarl.otp_service');

    /** @var \Drupal\dwarl\UserService $userService */
    $userService = $container->get('dwarl.user_service');

    /** @var \Drupal\dwarl\PasskeyStorageInterface $passkeyStorage */
    $passkeyStorage = $container->get('dwarl.passkey_storage');

    return new self(
      $requestStack->getCurrentRequest(),
      $anonymousStore,
      $logger,
      $jwtAuth,
      $privateTempStoreFactory,
      $requestOptions,
      $otpService,
      $userService,
      $passkeyStorage,
    );
  }

  /**
   * Set user password during registration.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setPassword(): JsonResponse {
    $password = $this->getPasswordFromRequest();
    $user = $this->anonymousStore->getUser();
    $this->userService->validateUserContactMethod($user);
    $user->setPassword($password)->activate()->save();

    // Clear store now registration is complete.
    $this->anonymousStore->expire();

    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();

    $context = ['@userId' => $user->id()];
    $this->logger->info('Set user @userId password', $context);

    return new JsonResponse(['token' => $token]);
  }

  /**
   * Authenticate user password and return JWT access token.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnException
   * @throws \JsonException
   */
  public function authenticate(): JsonResponse {
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadRequiredByUserHandle($userHandle);
    $password = $this->getPasswordFromRequest();
    $this->userService->validateUserPassword($user, $password);

    // Check if user has a passkey.
    $passkeys = $this->passkeyStorage->getUserPasskeys((int) $user->id());
    if ($passkeys) {
      // Send request options to validate passkey.
      $this->anonymousStore->init()->setUser($user);
      $this->anonymousStore->setRealm('login');
      $email = $user->getEmail();
      $data = [
        'requestOptions' => $this->requestOptions->getRequestOptions($email),
      ];
      return $this->anonymousStore->getSessionIdResponse($data);
    }

    // Return access token.
    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();
    $this->logger->info('jwt: ' . $token);
    return new JsonResponse(['token' => $token]);
  }

  /**
   * Request set password otp.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   *
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   */
  public function requestChangePasswordOtp(): JsonResponse {
    $user = $this->userService->currentUser();
    $password = $this->getFromRequest('password');
    $methodStr = $this->getRequiredString('method');
    $method = $this->otpService->validateMethod($methodStr);

    // Validate user password if they have one.
    if ($user->getPassword()) {
      $this->userService->validateUserPassword($user, $password);
    }

    $this->otpService->generateOtp($this->tempStore, $user, OtpSendReason::UpdatePassword, $method);
    return new JsonResponse();
  }

  /**
   * Request a set password otp.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   */
  public function requestSetPasswordOtp(): JsonResponse {
    $user = $this->userService->currentUser();
    if ($user->getPassword()) {
      $message = 'User already has password. Use change password endpoint.';
      throw new UserHasPasswordException($message);
    }
    return $this->requestChangePasswordOtp();
  }

  /**
   * Set password with an otp.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function setPasswordOtp(): JsonResponse {
    $this->otpService->validateOtp($this->tempStore);
    $user = $this->userService->currentUser();
    $password = $this->getPasswordFromRequest();
    $user->setPassword($password);
    $user->save();
    return new JsonResponse();
  }

}
