<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\dwarl\Exception\MissingFieldException;
use Drupal\dwarl\Exception\NoPasskeysException;
use Drupal\dwarl\Exception\UserHandleOrCodeInvalidException;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl\Util\CodeGenerator;
use Drupal\dwarl\WebAuthn\AuthenticateRequestInterface;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Dwarl recovery controller.
 */
class RecoveryController extends DwarlControllerBase {


  /**
   * Dwarl private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  private PrivateTempStore $tempStore;

  /**
   * DwarlRecoveryController constructor.
   */
  public function __construct(
    private readonly Request $request,
    private readonly RequestOptionsInterface $requestOptions,
    private readonly AuthenticateRequestInterface $authenticateRequest,
    private readonly CodeGenerator $codeGenerator,
    private readonly PasswordInterface $password,
    private readonly JwtAuth $jwtAuth,
    private readonly LoggerInterface $logger,
    private readonly UserService $userService,
    private readonly PasskeyStorageInterface $passkeyStorage,
  ) {
    parent::__construct($this->request, $logger);
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\WebAuthn\RequestOptionsInterface $requestOptions */
    $requestOptions = $container->get('dwarl.request_options');

    /** @var \Drupal\dwarl\WebAuthn\AuthenticateRequestInterface $authenticateRequest */
    $authenticateRequest = $container->get('dwarl.authenticate_request');

    /** @var \Drupal\dwarl\Util\CodeGenerator $codeGenerator */
    $codeGenerator = \Drupal::service('dwarl.code_generator');

    /** @var \Drupal\jwt\Authentication\Provider\JwtAuth $jwtAuth */
    $jwtAuth = $container->get('jwt.authentication.jwt');

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = \Drupal::service('logger.channel.dwarl');

    /** @var \Drupal\Core\Password\PasswordInterface $password */
    $password = $container->get('password');

    /** @var \Drupal\dwarl\UserService $userService */
    $userService = $container->get('dwarl.user_service');

    /** @var \Drupal\dwarl\PasskeyStorageInterface $passkeyStorage */
    $passkeyStorage = $container->get('dwarl.passkey_storage');

    return new self(
      $requestStack->getCurrentRequest(),
      $requestOptions,
      $authenticateRequest,
      $codeGenerator,
      $password,
      $jwtAuth,
      $logger,
      $userService,
      $passkeyStorage,
    );
  }

  /**
   * Request options for generate recovery code request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnException
   * @throws \JsonException
   */
  public function getRequestOption(): JsonResponse {
    $user = $this->userService->currentUser();
    $passkeys = $this->passkeyStorage->getUserPasskeys((int) $user->id());
    if (!$passkeys) {
      throw new NoPasskeysException();
    }
    $this->validateHashFieldPresent($user);
    return new JsonResponse([
      'requestOptions' => $this->requestOptions->getRequestOptions($user->getEmail()),
    ]);
  }

  /**
   * Generate account recover code.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   * @throws \Random\RandomException
   */
  public function generateRecoveryCode(): JsonResponse {
    $authenticationRequest = $this->getRequiredString('authentication_request');
    $user = $this->userService->currentUser();
    $this->validateHashFieldPresent($user);
    $email = $user->getEmail();
    $this->authenticateRequest->execute($email, $authenticationRequest);

    $code = $this->codeGenerator->generateCode(128, CodeGenerator::KEYSPACE_STANDARD);
    $hash = $this->password->hash($code);
    $user->set('field_recovery_code_hash', $hash);
    $user->save();
    return new JsonResponse(['recovery_code' => $code]);
  }

  /**
   * Recover account.
   *
   * Provide user handle and recovery code to get a JWT token.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   */
  public function recoverAccount(): JsonResponse {
    $recoveryCode = $this->getRequiredString('recovery_code');
    $userHandle = $this->getRequiredString('user_handle');
    $user = $this->userService->loadByUserHandle($userHandle);
    $validCode = FALSE;
    if ($user) {
      $this->validateHashFieldPresent($user);
      $hash = $user->get('field_recovery_code_hash')->getString();
      $validCode = $this->password->check($recoveryCode, $hash);
    }
    if ($validCode !== TRUE) {
      throw new UserHandleOrCodeInvalidException();
    }

    user_login_finalize($user);
    $token = $this->jwtAuth->generateToken();

    $context = ['@uid' => $user->id()];
    $this->logger->info('User @uid account recovery code used. ', $context);

    return new JsonResponse(['token' => $token]);
  }

  /**
   * Validate user has field_recovery_code_hash.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   */
  private function validateHashFieldPresent(UserInterface $user): void {
    if ($user->hasField('field_recovery_code_hash') === FALSE) {
      throw new MissingFieldException('recovery code hash', 'user');
    }
  }

}
