<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\Component\Utility\EmailValidatorInterface;
use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\Exception\UserSaveException;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\Register;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\WebAuthn\AuthenticateRequestInterface;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Register controller.
 */
class RegisterController extends DwarlControllerBase {

  /**
   * Dwarl private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  private PrivateTempStore $tempStore;

  /**
   * User controller constructor.
   */
  public function __construct(
    private readonly Request $request,
    private readonly AnonymousStoreInterface $anonymousStore,
    private readonly LoggerInterface $logger,
    PrivateTempStoreFactory $privateTempStoreFactory,
    private readonly OtpServiceInterface $otpService,
    private readonly Register $register,
  ) {
    parent::__construct($this->request, $logger);
    $this->tempStore = $privateTempStoreFactory->get('dwarl');
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $anonymousStore */
    $anonymousStore = $container->get('dwarl.anonymous_store');

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = \Drupal::service('logger.channel.dwarl');

    /** @var \Drupal\Core\TempStore\PrivateTempStoreFactory $privateTempStoreFactory */
    $privateTempStoreFactory = $container->get('tempstore.private.one_hour');

    /** @var \Drupal\dwarl\Otp\OtpServiceInterface $otpService */
    $otpService = $container->get('dwarl.otp_service');

    /** @var \Drupal\dwarl\Register $register */
    $register = $container->get('dwarl.register');

    return new self(
      $requestStack->getCurrentRequest(),
      $anonymousStore,
      $logger,
      $privateTempStoreFactory,
      $otpService,
      $register,
    );
  }

  /**
   * Register with email.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   * @throws \Random\RandomException
   */
  public function registerEmail(): JsonResponse {
    $email = $this->getRequiredString('email');
    $sessionId = $this->register->registerEmail($email);
    return $this->sessionIdResponse($sessionId);
  }

  /**
   * Verify phone or email.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\UserSaveException
   */
  public function verify(): JsonResponse {
    $this->otpService->validateOtp($this->anonymousStore);
    $user = $this->anonymousStore->getUser();
    $method = $this->anonymousStore->getUserHandleType();
    $this->validateContactMethod($method, $user);
    $this->anonymousStore->resetExpiry();
    return new JsonResponse();
  }

  /**
   * Validate user contact method.
   *
   * Adds either the validated_email or validated_phone phone role.
   * Sets preferred contact method if not already set.
   *
   * @throws \Drupal\dwarl\Exception\UserSaveException
   */
  private function validateContactMethod(ContactMethod $method, UserInterface $user): void {
    $context = ['@uid' => $user->id()];
    $userNeedsSave = FALSE;

    if ($method === ContactMethod::Email && $user->hasRole('validated_email') === FALSE) {
      $user->addRole('validated_email');
      $this->logger->debug('Added validated email role for user @uid', $context);
      if ($user->get('field_preferred_contact_method')->isEmpty()) {
        $user->set('field_preferred_contact_method', 'email');
      }
      $userNeedsSave = TRUE;
    }
    elseif ($method === ContactMethod::Phone && $user->hasRole('validated_phone') === FALSE) {
      $user->addRole('validated_phone');
      $this->logger->debug('Added validated phone role for user @uid', $context);
      if ($user->get('field_preferred_contact_method')->isEmpty()) {
        $user->set('field_preferred_contact_method', 'phone');
      }
      $userNeedsSave = TRUE;
    }

    if ($userNeedsSave) {
      try {
        $user->save();
        $this->logger->debug('Saved user @uid', $context);
      }
      catch (\Exception $exception) {
        throw new UserSaveException('Failed save user after adding role.', $exception);
      }
    }
  }

}
