<?php

declare(strict_types=1);

namespace Drupal\dwarl\Controller;

use Drupal\Component\Utility\EmailValidatorInterface;
use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\Exception\BadRequestException;
use Drupal\dwarl\Exception\BadStatusException;
use Drupal\dwarl\Exception\InvalidContactMethodException;
use Drupal\dwarl\Exception\InvalidEmailAddressException;
use Drupal\dwarl\Exception\InvalidPasswordException;
use Drupal\dwarl\Exception\NoAuthMethodException;
use Drupal\dwarl\Exception\NoValidatedEmailException;
use Drupal\dwarl\Exception\NoValidatedPhoneException;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl\Util\PhoneNumber;
use Drupal\dwarl\WebAuthn\AuthenticateRequestInterface;
use Drupal\dwarl\WebAuthn\RequestOptionsInterface;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Dwarl user controller.
 */
final class UserController extends DwarlControllerBase {

  /**
   * Dwarl private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  private PrivateTempStore $tempStore;

  /**
   * User controller constructor.
   */
  public function __construct(
    private readonly Request $request,
    private readonly PasskeyStorageInterface $passkeyStorage,
    private readonly RequestOptionsInterface $requestOptions,
    private readonly PasswordInterface $password,
    PrivateTempStoreFactory $privateTempStoreFactory,
    private readonly AuthenticateRequestInterface $authenticateRequest,
    private readonly EmailValidatorInterface $emailValidator,
    private readonly OtpServiceInterface $otpService,
    LoggerInterface $logger,
  ) {
    parent::__construct($this->request, $logger);
    $this->tempStore = $privateTempStoreFactory->get('dwarl');
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container): self {

    /** @var \Symfony\Component\HttpFoundation\RequestStack $requestStack */
    $requestStack = $container->get('request_stack');

    /** @var \Drupal\dwarl\PasskeyStorageInterface $passkeyStorage */
    $passkeyStorage = $container->get('dwarl.passkey_storage');

    /** @var \Drupal\dwarl\WebAuthn\RequestOptionsInterface $requestOptions */
    $requestOptions = $container->get('dwarl.request_options');

    /** @var \Drupal\Core\Password\PasswordInterface $password */
    $password = $container->get('password');

    /** @var \Drupal\Core\TempStore\PrivateTempStoreFactory $privateTempStoreFactory */
    $privateTempStoreFactory = $container->get('tempstore.private.one_hour');

    /** @var \Drupal\dwarl\WebAuthn\AuthenticateRequestInterface $authenticateRequest */
    $authenticateRequest = $container->get('dwarl.authenticate_request');

    /** @var \Drupal\Component\Utility\EmailValidatorInterface $emailValidator */
    $emailValidator = $container->get('email.validator');

    /** @var \Drupal\dwarl\Otp\OtpServiceInterface $otpService */
    $otpService = $container->get('dwarl.otp_service');

    /** @var \Psr\Log\LoggerInterface $logger */
    $logger = $container->get('logger.channel.dwarl');

    return new self(
      $requestStack->getCurrentRequest(),
      $passkeyStorage,
      $requestOptions,
      $password,
      $privateTempStoreFactory,
      $authenticateRequest,
      $emailValidator,
      $otpService,
      $logger,
    );
  }

  /**
   * Return information about the current user.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The user information.
   */
  public function userInfo(): JsonResponse {

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());

    $email = $user->getEmail();
    $email = str_ends_with($email, '@dwarl.com') ? '' : $email;

    $preferredContactMethod = $user->hasField('field_preferred_contact_method')
      ? $user->get('field_preferred_contact_method')->getString()
      : NULL;

    $phone = $user->hasField('field_phone')
      ? $user->get('field_phone')->getValue()
      : NULL;
    $phone = $phone[0]['value'] ?? '';

    return new JsonResponse([
      'id' => $user->id(),
      'email' => $email,
      'phone' => $phone,
      'preferredContactMethod' => $preferredContactMethod,
      'hasPassword' => $user->getPassword() !== NULL,
      'passkeys' => $this->passkeyStorage->getPasskeys((int) $user->id()),
    ]);
  }

  /**
   * Set email.
   *
   * If user has a passkey request options are returned.
   * If no passkey and password validates an OTP is sent.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnException
   * @throws \JsonException
   */
  public function setContactMethod(): JsonResponse {

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());
    $passkeys = $this->passkeyStorage->getPasskeys((int) $user->id());

    return $passkeys
      ? $this->setContactMethodPasskeyVerificationRequiredResponse($user)
      : $this->setContactMethodPasswordRequiredResponse($user);
  }

  /**
   * Check if username is taken.
   */
  public function checkUsername(): JsonResponse {
    $username = $this->getRequiredString('username');
    $existing = \Drupal::entityQuery('user')
      ->accessCheck(FALSE)
      ->condition('name', $username)
      ->range(0, 1)
      ->execute();
    return new JsonResponse(['status' => $existing ? 'taken' : 'available']);
  }

  /**
   * Set username.
   */
  public function setUsername(): JsonResponse {
    $username = $this->getRequiredString('username');

    $existing = \Drupal::entityQuery('user')
      ->accessCheck(FALSE)
      ->condition('name', $username)
      ->range(0, 1)
      ->execute();

    if ($existing) {
      throw new BadRequestException('Username already exists');
    }

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());

    try {
      $user->setUsername($username);
      $user->save();
    }
    catch (\Exception $e) {
      throw new BadRequestException('User save failed');
    }

    return new JsonResponse();

  }

  /**
   * Set contact method build password required response.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\dwarl\Exception\NoAuthMethodException
   */
  private function setContactMethodPasswordRequiredResponse(UserInterface $user): JsonResponse {
    $password = $user->getPassword();
    if ($password === NULL) {
      throw new NoAuthMethodException();
    }

    $this->tempStore->set('setEmailMethod', 'password');
    return new JsonResponse(['passwordRequired' => TRUE]);
  }

  /**
   * Set contact method build passkey verification required response.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnException
   * @throws \JsonException
   */
  private function setContactMethodPasskeyVerificationRequiredResponse(UserInterface $user): JsonResponse {
    // This user has a passkey so request verification before updating email.
    $this->tempStore->set('setEmailMethod', 'passkey');
    return new JsonResponse([
      'requestOptions' => $this->requestOptions->getRequestOptions($user->getEmail()),
    ]);
  }

  /**
   * Validate set email authentication request and sent OTP.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  public function setContactMethodRequestOtp(): JsonResponse {

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());

    $method = $this->tempStore->get('setEmailMethod');
    $userHandle = $this->getRequiredString('user_handle');
    [$email, $phone] = UserService::getEmailPhone($userHandle);
    $otpSendMethod = $this->setContactMethodGetOtpSendMethod($email, $phone);
    $this->setContactMethodAuthenticateRequest($method, $user);
    $reason = $email ? OtpSendReason::UpdateEmail : OtpSendReason::UpdatePhone;
    $this->otpService->generateOtp($this->tempStore, $user, $reason, $otpSendMethod, $userHandle);
    return new JsonResponse();
  }

  /**
   * Get OTP send method amd update the temp store.
   *
   * @param string|null $email
   *   The email.
   * @param string|null $phone
   *   The phone.
   *
   * @return \Drupal\dwarl\ContactMethod
   *   The OTP send method to use.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   */
  private function setContactMethodGetOtpSendMethod(?string $email, ?string $phone): ContactMethod {
    if ($email) {
      if ($this->emailValidator->isValid($email) === FALSE) {
        throw new InvalidEmailAddressException();
      }
      $this->tempStore->set('email', $email);
      $this->tempStore->set('phone', NULL);
      return ContactMethod::Email;
    }

    $formatted = PhoneNumber::getFormattedPhoneNumber($phone);
    $this->tempStore->set('phone', $formatted);
    $this->tempStore->set('email', NULL);
    return ContactMethod::Phone;
  }

  /**
   * Set contact method check passkey or password.
   *
   * @param string $method
   *   The method.
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnInvalidArgumentException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnUnexpectedValueException
   * @throws \Drupal\public_key_credential_source\Exception\PublicKeyCredentialSourceWebauthnValidationFailedException
   * @throws \Drupal\webauthn_jwt\Exception\WebauthnJwtException
   */
  private function setContactMethodAuthenticateRequest(string $method, UserInterface $user): void {
    if ($method === 'passkey') {
      $authenticationRequest = $this->getRequiredString('authentication_request');
      $userCurrentEmail = $user->getEmail();
      $this->authenticateRequest->execute($userCurrentEmail, $authenticationRequest);
    }
    elseif ($method === 'password') {
      $password = $this->getRequiredString('password');
      $valid = $this->password->check($password, $user->getPassword());
      if (!$valid) {
        throw new InvalidPasswordException();
      }
    }
    else {
      // User must have visited the previous endpoints first to avoid this.
      throw new BadStatusException('Bad method');
    }
  }

  /**
   * Validate set email OTP and update user email.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\dwarl\Exception\InvalidOtpException
   * @throws \Drupal\dwarl\Exception\OtpBlockedException
   * @throws \Drupal\dwarl\Exception\BadRequestException
   */
  public function setContactMethodConfirm(): JsonResponse {
    $this->otpService->validateOtp($this->tempStore);
    $email = $this->tempStore->get('email');
    $phone = $this->tempStore->get('phone');

    if ($email && $phone) {
      throw new BadRequestException('Can only provide email or phone');
    }

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());

    if ($email) {
      // @todo if user has existing email send message to say it been changed.
      $user->setEmail($email);
      $user->addRole('validated_email');
    }
    if ($phone) {
      // @todo if user has existing phone send message to say it been changed.
      $user->set('field_phone', $phone);
      $user->addRole('validated_phone');
    }

    $user->save();

    return new JsonResponse();
  }

  /**
   * Set preferred contact method.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\dwarl\Exception\InvalidContactMethodException
   * @throws \Drupal\dwarl\Exception\NoValidatedEmailException
   * @throws \Drupal\dwarl\Exception\NoValidatedPhoneException
   */
  public function setPreferredContactMethod(): JsonResponse {
    $method = $this->getRequiredString('method');

    if (!in_array($method, ['email', 'phone'])) {
      throw new InvalidContactMethodException();
    }

    /** @var \Drupal\user\UserInterface $user */
    $user = User::load($this->currentUser()->id());

    $validatedEmail = $user->hasRole('validated_email');
    $validatedPhone = $user->hasRole('validated_phone');

    if ($method === 'email' && ($user->getEmail() === NULL || !$validatedEmail)) {
      throw new NoValidatedEmailException();
    }

    if ($method === 'phone' && ($user->get('field_phone')->isEmpty() || $validatedPhone)) {
      throw new NoValidatedPhoneException();
    }

    if ($user->hasField('field_preferred_contact_method')) {
      $user->set('field_preferred_contact_method', $method);
      $user->save();
    }

    return new JsonResponse();
  }

}
