<?php

declare(strict_types=1);

namespace Drupal\dwarl\Event;

use Drupal\Component\EventDispatcher\Event;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\Exception\NoValidatedContactMethodException;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\user\UserInterface;

/**
 * Send OTP event.
 *
 * Reasons why an OTP might be sent:
 *  1. User register - confirm email or phone.
 *  2. Change email or phone.
 *  3. Set new email or phone.
 *  4. Forgot password.
 *  5. Change password.
 *  6. Lost passkey.
 */
class SendOtpEvent extends Event {

  public const SEND_OTP = 'send_otp_event';

  /**
   * Email override.
   *
   * @var string|null
   */
  private ?string $emailOverride;

  /**
   * Phone override.
   *
   * @var string|null
   */
  private ?string $phoneOverride;

  /**
   * SendOtpEvent constructor.
   *
   * @param string $otp
   *   The one time pass.
   * @param \Drupal\user\UserInterface $user
   *   The user.
   * @param \Drupal\dwarl\ContactMethod|null $method
   *   The send OTP method.
   * @param string|null $emailPhoneOverride
   *   A phone or email to send to OTP to, overriding the default value.
   */
  public function __construct(
    private readonly string $otp,
    private readonly UserInterface $user,
    private readonly OtpSendReason $reason,
    private readonly ?ContactMethod $method = NULL,
    ?string $emailPhoneOverride = NULL,
  ) {
    $isEmail = $emailPhoneOverride && str_contains($emailPhoneOverride, '@');
    $this->emailOverride = $isEmail ? $emailPhoneOverride : NULL;
    $this->phoneOverride = $isEmail ? NULL : $emailPhoneOverride;
  }

  /**
   * Get OTP.
   *
   * @return string
   *   The OTP.
   */
  public function getOtp(): string {
    return $this->otp;
  }

  /**
   * Get OTP send reason.
   *
   * @return \Drupal\dwarl\Otp\OtpSendReason
   *   The reason.
   */
  public function getReason(): OtpSendReason {
    return $this->reason;
  }

  /**
   * Get send OTP method.
   *
   * @return \Drupal\dwarl\ContactMethod
   *   The send OTP method.
   */
  public function getMethod(): ContactMethod {
    if ($this->method) {
      return $this->method;
    }

    $email = $this->user->getEmail();
    // Users who register with phone have their email set to <PHONE>@dwarl.com.
    $hasRealEmail = str_ends_with($email, '@dwarl.com') === FALSE;

    $phone = $this->user->hasField('field_phone')
      ? $this->user->get('field_phone')->getString()
      : NULL;

    $fieldPcm = $this->user->hasField('field_preferred_contact_method')
      ? $this->user->get('field_preferred_contact_method')->getString()
      : NULL;
    $preferredContactMethod = $fieldPcm ? ContactMethod::tryFrom($fieldPcm) : NULL;

    if ($hasRealEmail && $preferredContactMethod === ContactMethod::Email) {
      return ContactMethod::Email;
    }
    if ($phone && $preferredContactMethod === ContactMethod::Phone) {
      return ContactMethod::Phone;
    }
    if ($email) {
      return ContactMethod::Email;
    }
    if ($phone) {
      return ContactMethod::Phone;
    }

    throw new NoValidatedContactMethodException();
  }

  /**
   * Get user.
   *
   * @return \Drupal\user\UserInterface
   *   The user.
   */
  public function getUser(): UserInterface {
    return $this->user;
  }

  /**
   * Get email to send to.
   *
   * @return string|null
   *   The email to send to.
   */
  public function getEmail(): ?string {
    return $this->emailOverride ?? $this->user->getEmail();
  }

  /**
   * Get phone to send to.
   *
   * @return string|null
   *   The phone to send to.
   */
  public function getPhone(): ?string {
    return $this->phoneOverride ?? $this->user->get('field_phone')->getValue()[0]['value'] ?? NULL;
  }

}
