<?php

declare(strict_types=1);

namespace Drupal\dwarl\Otp;

use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\Event\SendOtpEvent;
use Drupal\dwarl\Exception\FailedToGenerateOtpException;
use Drupal\dwarl\Exception\InvalidContactMethodException;
use Drupal\dwarl\Exception\InvalidOtpException;
use Drupal\dwarl\Exception\OtpBlockedException;
use Drupal\dwarl\Store\StoreInterface;
use Drupal\dwarl\Util\CodeGenerator;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * @todo Add class description.
 */
final class OtpService implements OtpServiceInterface {

  const MAX_ATTEMPTS = 3;
  const GENERATION_RATE_LIMIT = 120;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request|null
   */
  private ?Request $request;

  /**
   * Constructs an OtpService object.
   */
  public function __construct(
    private readonly CodeGenerator $codeGenerator,
    RequestStack $requestStack,
    private readonly EventDispatcherInterface $eventDispatcher,
    private readonly LoggerInterface $logger,
    private readonly PasswordInterface $password,
  ) {
    $this->request = $requestStack->getCurrentRequest();
  }

  /**
   * {@inheritdoc}
   */
  public function generateOtp(
    PrivateTempStore|StoreInterface $store,
    UserInterface $user,
    OtpSendReason $reason,
    ?ContactMethod $otpSendMethod = NULL,
    ?string $emailPhoneOverride = NULL,
  ): void {

    try {
      $otp = $this->generateAndSaveOtp($store);
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception, $reason->name . ': ' . $exception->getMessage());
      throw new FailedToGenerateOtpException();
    }

    $sendOtpEvent = new SendOtpEvent($otp, $user, $reason, $otpSendMethod, $emailPhoneOverride);
    $this->eventDispatcher->dispatch($sendOtpEvent, SendOtpEvent::SEND_OTP);

    // @todo Don't store plaintext OTP in the db.
    $context = [
      '@event' => $reason->value,
      '@userId' => $user->id(),
      '@otp' => $otp,
    ];
    $this->logger->info('Otp for @event event for user @userId: @otp', $context);
  }

  /**
   * {@inheritdoc}
   */
  public function validateOtp(PrivateTempStore|StoreInterface $store, ?string $otp = NULL): void {
    $otp = $otp ?? $this->request->get('otp');
    if (empty($otp)) {
      throw new InvalidOtpException('OTP missing from request');
    }

    $attempts = $store->get('otp_attempts');
    if ($attempts === NULL || $attempts > self::MAX_ATTEMPTS) {
      throw new OtpBlockedException('Attempts: ' . $attempts);
    }

    $expires = $store->get('otp_expires');
    if ($expires === NULL || time() > $expires) {
      throw new OtpBlockedException();
    }

    // Check otp matches saved hash.
    $savedHash = $store->get('otp');
    if ($this->password->check($otp, $savedHash) === FALSE) {
      throw new InvalidOtpException('Bad OTP');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validateMethod(?string $method): ContactMethod {
    try {
      $contactMethod = ContactMethod::from($method);
    }
    catch (\Exception) {
      throw new InvalidContactMethodException();
    }
    return $contactMethod;
  }

  /**
   * Generate OTP and save hash in local store.
   *
   * @param \Drupal\Core\TempStore\PrivateTempStore|\Drupal\dwarl\Store\StoreInterface $store
   *   The store.
   *
   * @return string
   *   The OTP.
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Random\RandomException
   */
  public function generateAndSaveOtp(PrivateTempStore|StoreInterface $store): string {
    $otp = $this->codeGenerator->generateCode();
    $hash = $this->password->hash($otp);
    $store->set('otp', $hash);
    $store->set('otp_attempts', 0);
    $store->set('otp_expires', strtotime('+ 30 minutes'));
    return $otp;
  }

}
