<?php

declare(strict_types=1);

namespace Drupal\dwarl;

use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\Exception\MoreThanOnePasskeyException;
use Drupal\dwarl\Exception\NoPasskeysException;
use Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource;
use Drupal\public_key_credential_source\PublicKeyCredentialSourceInterface;
use Psr\Log\LoggerInterface;

/**
 * Class for loading passkeys.
 */
final class PasskeyStorage implements PasskeyStorageInterface {

  /**
   * Passkey storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  private EntityStorageInterface $keyStorage;

  /**
   * Constructs a PasskeyStorage object.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerInterface $logger,
  ) {
    $this->keyStorage = $entityTypeManager->getStorage('public_key_credential_source');
  }

  /**
   * {@inheritdoc}
   */
  public function getPasskeys(int $userId): array {
    $ids = $this->keyStorage->getQuery()
      ->accessCheck()
      ->condition('uid', $userId)
      ->execute();

    /** @var \Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource[] $entities */
    $entities = $this->keyStorage->loadMultiple($ids);

    $passkeys = [];
    foreach ($entities as $entity) {
      $passkeys[] = [
        'id' => $entity->id(),
        'created' => $entity->get('created')->getString(),
        'last_used' => $entity->get('last_used')->getString(),
        'created_device' => $entity->get('created_device')->getString(),
        'last_used_device' => $entity->get('last_used_device')->getString(),
      ];
    }

    return $passkeys;
  }

  /**
   * {@inheritDoc}
   */
  public function deleterForUser(int $userId): void {
    $ids = $this->keyStorage->getQuery()
      ->accessCheck(FALSE)
      ->condition('uid', $userId)
      ->execute();
    if ($ids) {
      $entities = $this->keyStorage->loadMultiple($ids);
      $this->keyStorage->delete($entities);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function getUserPasskeys(int $userId): array {
    return $this->getPasskeys($userId);
  }

  /**
   * {@inheritDoc}
   */
  public function getUserSinglePasskey(int $userId): PublicKeyCredentialSourceInterface {
    $passkeys = $this->getPasskeys($userId);

    // If user has no passkeys return an error.
    if (!$passkeys) {
      throw new NoPasskeysException();
    }

    // If multiple passkeys get user to authenticate with a different one.
    if (count($passkeys) > 1) {
      throw new MoreThanOnePasskeyException();
    }

    return PublicKeyCredentialSource::load($passkeys[0]['id']);
  }

  /**
   * {@inheritDoc}
   */
  public function deletePasskey(PublicKeyCredentialSourceInterface $publicKeyCredentialSource): void {
    try {
      $publicKeyCredentialSource->delete();
    }
    catch (\Throwable $exception) {
      // It's not ideal the token failed to delete, but let the user carry on.
      Error::logException($this->logger, $exception);
    }
  }

}
