<?php

declare(strict_types=1);

namespace Drupal\dwarl;

use Drupal\Component\Utility\EmailValidatorInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\Access\RegistrationAccessCheck;
use Drupal\dwarl\Exception\InvalidEmailAddressException;
use Drupal\dwarl\Otp\OtpSendReason;
use Drupal\dwarl\Otp\OtpServiceInterface;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\Util\PhoneNumber;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;

/**
 * Register service for creating user accounts.
 */
class Register implements RegisterInterface {

  /**
   * User storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  private EntityStorageInterface $userStorage;

  /**
   * Constructs Register service.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerInterface $logger,
    private readonly EmailValidatorInterface $emailValidator,
    private readonly AnonymousStoreInterface $anonymousStore,
    private readonly OtpServiceInterface $otpService,
    private readonly UserServiceInterface $userService,
    private readonly PasskeyStorageInterface $passkeyStorage,
  ) {
    $this->userStorage = $entityTypeManager->getStorage('user');
  }

  /**
   * {@inheritDoc}
   */
  public function registerEmail(string $email): string {

    // Validate email.
    if ($email && $this->emailValidator->isValid($email) === FALSE) {
      throw new InvalidEmailAddressException();
    }

    // Load user from email.
    $user = $this->userService->loadByUserHandle($email);

    return $this->register($user, $email, NULL);
  }

  /**
   * {@inheritDoc}
   */
  public function registerPhone(string $phone, string $country): string {
    $formatted = PhoneNumber::getFormattedPhoneNumber($phone, $country);

    // Load user from phone.
    $user = $this->userService->loadByUserHandle($formatted);

    return $this->register($user, NULL, $formatted, $country);
  }

  /**
   * Register account.
   *
   * @param \Drupal\user\UserInterface|null $user
   *   Maybe a user, maybe not.
   * @param string|null $email
   *   Maybe an email, maybe not.
   * @param string|null $phone
   *   Maybe a phone, maybe not.
   * @param string|null $country
   *   The country for the phone number.
   *
   * @return string
   *   The session id.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\StoreExpiredException
   * @throws \Drupal\dwarl\Exception\FailedToGenerateOtpException
   */
  private function register(
    ?UserInterface $user,
    ?string $email,
    ?string $phone,
    ?string $country = NULL,
  ): string {
    $userHandle = $email ?? $phone;

    if ($user) {
      // Handle when user already exists.
      $sendOtp = $this->handleExistingUser($user, $userHandle);
    }
    else {
      // Create new user.
      $sendOtp = TRUE;
      $user = $this->createUser($userHandle, $email, $phone, $country);
    }

    if ($user->hasField('field_preferred_contact_method')) {
      $user->set('field_preferred_contact_method', $email ? 'email' : 'phone');
    }

    // Generate OTP if needed.
    $isEmail = $phone === NULL;
    $userHandleType = $isEmail ? ContactMethod::Email : ContactMethod::Phone;

    $this->anonymousStore
      ->init()
      ->setRealm(RegistrationAccessCheck::REALM)
      ->setUser($user)
      ->setUserHandleType($userHandleType);

    if ($sendOtp) {
      $this->otpService->generateOtp($this->anonymousStore, $user, OtpSendReason::Register);
    }

    // Return session id.
    return $this->anonymousStore->getSessionId();
  }

  /**
   * Handle existing user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   * @param string $userHandle
   *   The user handle.
   *
   * @return bool
   *   Should the OPT be sent?
   */
  private function handleExistingUser(UserInterface $user, string $userHandle): bool {
    // If user doesn't have a password or a passkey allow them to continue.
    $hasPassword = $user->getPassword() !== NULL;
    $passKeys = $this->passkeyStorage->getPasskeys((int) $user->id());
    $hasPasskey = count($passKeys) > 0;

    if ($hasPassword || $hasPasskey) {
      // This user can already login so cannot continue registration.
      // Don't send an OTP, but do return a session id so we don't leak that
      // this user already exists.
      $args = ['@handle' => $userHandle];
      $this->logger
        ->warning('Registration attempt for existing user @handle', $args);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * {@inheritDoc}
   */
  public function createUser(string $userHandle, ?string $email, ?string $phone, ?string $country = NULL): UserInterface {

    /** @var \Drupal\user\UserInterface $user */
    $user = $this->userStorage->create();

    $user->setUsername($userHandle);

    /** @var \Drupal\phone_number\PhoneNumberUtil $phoneUtil */
    $phoneUtil = \Drupal::service('phone_number.util');

    if ($email) {
      $user->setEmail($email);
    }
    if ($phone) {
      $phoneEntity = $phoneUtil->getPhoneNumber($phone);
      if ($phoneEntity === NULL) {
        throw new \RuntimeException('Invalid phone number.');
      }

      $user->set('field_phone', [
        'value' => $phone,
        'country' => $country,
        'local_number' => $phoneUtil->getLocalNumber($phoneEntity),
      ]);

      // Due to restrictions in public_key_credential_source users must
      // have an email to be able to create a passkey.
      /* @see \Drupal\public_key_credential_source\Services\PublicKeyCredentialSourceWebauthn::getRegistrationOptions */
      $email = str_replace([' ', '-'], '', $phone) . '@dwarl.com';
      $user->setEmail($email);
    }

    try {
      $user->save();
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception, 'Failed to save user.');
      throw new \RuntimeException('Could not save user.');
    }

    // Log new account creation.
    $args = ['@userHandle' => $userHandle];
    $this->logger
      ->info('Created new user account @userHandle', $args);

    return $user;
  }

}
