<?php

declare(strict_types=1);

namespace Drupal\dwarl\Store;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\Exception\BadStatusException;
use Drupal\dwarl\Exception\DwarlStoreException;
use Drupal\dwarl\Exception\NoRequestException;
use Drupal\dwarl\Exception\StoreExpiredException;
use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Session\SessionInterface;

/**
 * Store for unauthenticated users.
 */
final class AnonymousStore implements AnonymousStoreInterface {

  /**
   * How many seconds does the store keep things for?
   *
   * Half an hour.
   */
  private const VALIDITY_PERIOD = 1800;

  private const SESSION_USER_KEY = 'core.tempstore.private.owner';

  private const TEMP_STORE_KEY = 'dwarl';

  /**
   * The current session.
   *
   * @var \Symfony\Component\HttpFoundation\Session\SessionInterface
   */
  private SessionInterface $session;

  /**
   * The session id.
   *
   * @var string
   */
  private string $sessionId;

  /**
   * The private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  public PrivateTempStore $store;

  /**
   * Constructs a Store object.
   *
   * @throws \Drupal\dwarl\Exception\NoRequestException
   */
  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerChannelInterface $logger,
    RequestStack $requestStack,
    PrivateTempStoreFactory $privateTempStoreFactory,
  ) {
    $request = $requestStack->getCurrentRequest();
    if ($request === NULL) {
      throw new NoRequestException();
    }
    $this->session = $request->getSession();
    $this->store = $privateTempStoreFactory->get(self::TEMP_STORE_KEY);
  }

  /**
   * {@inheritdoc}
   */
  public function init(?string $sessionId = NULL): AnonymousStoreInterface {

    if ($sessionId) {
      // Restore an existing session. Check expiry.
      $this->sessionId = $sessionId;
      $this->session->set(self::SESSION_USER_KEY, $this->sessionId);
      $this->logger->debug('Restore store with ' . $this->sessionId);

      if ($this->isExpired()) {
        throw new StoreExpiredException();
      }
    }
    else {
      // Start a new session. Set expiry.
      $this->sessionId = Crypt::randomBytesBase64(6);
      $this->session->set(self::SESSION_USER_KEY, $this->sessionId);
      $this->logger->debug('Init store with ' . $this->sessionId);
    }

    return $this->resetExpiry();
  }

  /**
   * {@inheritdoc}
   */
  public function getSessionId(): string {
    return $this->sessionId;
  }

  /**
   * {@inheritdoc}
   */
  public function getSessionIdResponse(mixed $data = NULL): JsonResponse {
    $response = new JsonResponse($data);
    $response->headers->set('session-id', $this->getSessionId());
    $response->headers->set('Access-Control-Expose-Headers', ['session-id']);
    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function setUser(UserInterface $user): AnonymousStoreInterface {
    return $this->set('user_id', $user->id());
  }

  /**
   * {@inheritdoc}
   */
  public function getUser(): UserInterface {
    $userId = $this->store->get('user_id');
    if (!$userId) {
      throw new DwarlStoreException('User id not available.');
    }
    try {
      $userStorage = $this->entityTypeManager->getStorage('user');
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception);
      throw new \RuntimeException('User storage not available.');
    }

    $user = $userStorage->load($userId);
    if (!$user instanceof UserInterface) {
      throw new DwarlStoreException('User not found.');
    }
    return $user;
  }

  /**
   * {@inheritdoc}
   */
  public function setUserHandleType(ContactMethod $userHandleType): AnonymousStoreInterface {
    $this->set('user_handle_type', $userHandleType);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getUserHandleType(): ContactMethod {
    $userHandleType = $this->store->get('user_handle_type');
    if (!$userHandleType) {
      throw new DwarlStoreException('User handle type not available.');
    }
    return $userHandleType;
  }

  /**
   * {@inheritdoc}
   */
  public function setOtp(string $otp): AnonymousStoreInterface {
    $this->set('otp', $otp);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getOtp(): string {
    $otp = $this->store->get('otp');
    if (!$otp) {
      throw new DwarlStoreException('No OTP in store.');
    }
    return $otp;
  }

  /**
   * {@inheritdoc}
   */
  public function resetExpiry(): AnonymousStoreInterface {
    $timestamp = strtotime('+ ' . self::VALIDITY_PERIOD . ' seconds');
    return $this->set('expires', $timestamp);
  }

  /**
   * {@inheritdoc}
   */
  public function isExpired(): bool {
    $expires = $this->store->get('expires');
    return $expires < time();
  }

  /**
   * {@inheritdoc}
   */
  public function expire(): void {
    $this->set('expires', time() - 1);
    $this->session->set(self::SESSION_USER_KEY, NULL);
  }

  /**
   * {@inheritdoc}
   */
  public function empty(): void {
    $this->logger->debug('Empty store with ' . $this->sessionId);
    unset($this->sessionId);
    $this->session->set(self::SESSION_USER_KEY, NULL);
    $this
      ->set('user_id', NULL)
      ->set('user_handle_type', NULL)
      ->set('otp', NULL)
      ->set('expires', NULL);
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): string {
    return $this->store->get('status') ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function validateStatus(string $requiredStatus): void {
    $status = $this->getStatus();
    if ($status !== $requiredStatus) {
      throw new BadStatusException();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setStatus(string $status): AnonymousStoreInterface {
    return $this->set('status', $status);
  }

  /**
   * {@inheritdoc}
   */
  public function setRealm(string $realm): AnonymousStoreInterface {
    return $this->set('realm', $realm);
  }

  /**
   * {@inheritdoc}
   */
  public function getRealm(): string {
    return $this->store->get('realm') ?? '';
  }

  /**
   * {@inheritDoc}
   */
  public function set(string $key, $value): AnonymousStoreInterface {
    try {
      $this->store->set($key, $value);
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception);
      throw new DwarlStoreException('Store not available');
    }
    return $this;
  }

  /**
   * {@inheritDoc}
   */
  public function get($key) {
    return $this->store->get($key);
  }

}
