<?php

declare(strict_types=1);

namespace Drupal\dwarl\Store;

use Drupal\dwarl\ContactMethod;
use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * Dwarl private temp store interface.
 */
interface AnonymousStoreInterface extends StoreInterface {

  /**
   * Setup the store.
   *
   * @param string|null $sessionId
   *   Session id. One is generated if not passed in.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\StoreExpiredException
   */
  public function init(?string $sessionId = NULL): self;

  /**
   * Get session id.
   *
   * @return string
   *   The session id.
   */
  public function getSessionId(): string;

  /**
   * Get json response with session id.
   *
   * @param mixed|null $data
   *   The response data.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The json response.
   */
  public function getSessionIdResponse(mixed $data = NULL): JsonResponse;

  /**
   * Set user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setUser(UserInterface $user): self;

  /**
   * Get user.
   *
   * @return \Drupal\user\UserInterface
   *   The user.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function getUser(): UserInterface;

  /**
   * Set user handle type.
   *
   * @param \Drupal\dwarl\ContactMethod $userHandleType
   *   The user handle type.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setUserHandleType(ContactMethod $userHandleType): self;

  /**
   * Get user handle type.
   *
   * @return \Drupal\dwarl\ContactMethod
   *   The user handle type.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function getUserHandleType(): ContactMethod;

  /**
   * Set OTP.
   *
   * @param string $otp
   *   The OTP.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setOtp(string $otp): self;

  /**
   * Get OTP.
   *
   * @return string
   *   THe OTP.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function getOtp(): string;

  /**
   * Reset expiry time.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function resetExpiry(): self;

  /**
   * Check if data is expired.
   *
   * @return bool
   *   Is the data expired?
   */
  public function isExpired(): bool;

  /**
   * Expire the store.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function expire(): void;

  /**
   * Empty the store.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function empty(): void;

  /**
   * Set status.
   *
   * @param string $status
   *   The status string.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setStatus(string $status): self;

  /**
   * Get status.
   *
   * @return string
   *   The status string.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function getStatus(): string;

  /**
   * Validate status matches.
   *
   * @param string $requiredStatus
   *   The required status string.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   * @throws \Drupal\dwarl\Exception\BadStatusException
   */
  public function validateStatus(string $requiredStatus): void;

  /**
   * Set realm.
   *
   * @param string $realm
   *   The realm.
   *
   * @return $this
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function setRealm(string $realm): self;

  /**
   * Get realm.
   *
   * @return string
   *   The realm.
   *
   * @throws \Drupal\dwarl\Exception\DwarlStoreException
   */
  public function getRealm(): string;

}
