<?php

declare(strict_types=1);

namespace Drupal\dwarl;

use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\Exception\DwarlPhoneNotEnabledException;
use Drupal\dwarl\Exception\InvalidPasswordException;
use Drupal\dwarl\Exception\NoValidatedContactMethodException;
use Drupal\dwarl\Exception\UserNotFoundException;
use Drupal\dwarl\Exception\UserSaveException;
use Drupal\dwarl\Util\PhoneNumber;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;

/**
 * Methods to load and save users.
 */
class UserService implements UserServiceInterface {

  /**
   * User storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  private EntityStorageInterface $userStorage;

  /**
   * Constructs a CreateUser object.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerInterface $logger,
    private readonly ModuleHandlerInterface $moduleHandler,
    private readonly AccountProxyInterface $currentUser,
    private readonly PasswordInterface $password,
  ) {
    $this->userStorage = $entityTypeManager->getStorage('user');
  }

  /**
   * {@inheritDoc}
   */
  public function currentUser(): UserInterface {
    return User::load($this->currentUser->id());
  }

  /**
   * {@inheritDoc}
   */
  public function loadByUserHandle(string $userHandle): ?UserInterface {
    [$email, $phone] = self::getEmailPhone($userHandle);

    if ($email) {
      $user = user_load_by_mail($email);
    }
    else {
      if ($this->moduleHandler->moduleExists('dwarl_phone') === FALSE) {
        throw new DwarlPhoneNotEnabledException();
      }

      $phoneFormatted = PhoneNumber::getFormattedPhoneNumber($phone);
      $users = $this->userStorage->loadByProperties(['field_phone' => $phoneFormatted]);
      $user = $users ? reset($users) : NULL;
    }

    return $user instanceof UserInterface ? $user : NULL;
  }

  /**
   * {@inheritDoc}
   */
  public static function getEmailPhone(string $userHandle): array {
    $email = str_contains($userHandle, '@') ? $userHandle : NULL;
    $phone = $email ? NULL : $userHandle;
    return [$email, $phone];
  }

  /**
   * {@inheritDoc}
   */
  public function loadRequiredByUserHandle(string $userHandle): UserInterface {
    $user = $this->loadByUserHandle($userHandle);
    if (!$user instanceof UserInterface) {
      throw new UserNotFoundException();
    }
    return $user;
  }

  /**
   * {@inheritDoc}
   */
  public function saveUser(UserInterface $user): void {
    try {
      $user->save();
    }
    catch (\Exception $exception) {
      Error::logException($this->logger, $exception);
      throw new UserSaveException();
    }
  }

  /**
   * {@inheritDoc}
   */
  public function loadByUuid(string $uuid): ?UserInterface {
    $users = $this->userStorage->loadByProperties(['uuid' => $uuid]);
    $user = $users ? reset($users) : NULL;
    return $user instanceof UserInterface ? $user : NULL;
  }

  /**
   * {@inheritDoc}
   */
  public function validateUserPassword(UserInterface $user, string $password): void {
    // If user has a password set then validate it.
    $hash = $user->getPassword();
    $valid = $hash && $this->password->check($password, $hash);
    if (!$valid) {
      throw new InvalidPasswordException();
    }
  }

  /**
   * {@inheritDoc}
   */
  public function validateUserContactMethod(UserInterface $user): void {
    $validatedEmail = $user->hasRole('validated_email');
    $validatedPhone = $user->hasRole('validated_phone');
    if (!$validatedEmail && !$validatedPhone) {
      throw new NoValidatedContactMethodException();
    }
  }

}
