<?php

declare(strict_types=1);

namespace Drupal\dwarl;

use Drupal\user\UserInterface;

/**
 * User service interface.
 */
interface UserServiceInterface {

  /**
   * Get the current user.
   *
   * @return \Drupal\user\UserInterface
   *   The user.
   */
  public function currentUser(): UserInterface;

  /**
   * Load a user.
   *
   * @param string $userHandle
   *   The user handle. Either email or phone.
   *
   * @return \Drupal\user\UserInterface|null
   *   The user or null if not found.
   *
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   */
  public function loadByUserHandle(string $userHandle): ?UserInterface;

  /**
   * Get email and phone from user handle.
   *
   * @param string $userHandle
   *   The user handle.
   *
   * @return array
   *   Array with email and phone.
   */
  public static function getEmailPhone(string $userHandle): array;

  /**
   * Load a user.
   *
   * @param string $userHandle
   *   The user handle. Either email or phone.
   *
   * @return \Drupal\user\UserInterface
   *   The user.
   *
   * @throws \Drupal\dwarl\Exception\UserNotFoundException
   * @throws \Drupal\dwarl\Exception\DwarlPhoneNotEnabledException
   */
  public function loadRequiredByUserHandle(string $userHandle): UserInterface;

  /**
   * Try and save user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @throws \Drupal\dwarl\Exception\UserSaveException
   */
  public function saveUser(UserInterface $user): void;

  /**
   * Load user by uuid.
   *
   * @param string $uuid
   *   The uuid.
   *
   * @return \Drupal\user\UserInterface|null
   *   The user or null if not found.
   */
  public function loadByUuid(string $uuid): ?UserInterface;

  /**
   * Validate the user's password.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   * @param string $password
   *   The password to validate.
   *
   * @throws \Drupal\dwarl\Exception\InvalidPasswordException
   */
  public function validateUserPassword(UserInterface $user, string $password): void;

  /**
   * Validate user has confirmed email or phone.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @throws \Drupal\dwarl\Exception\NoValidatedContactMethodException
   */
  public function validateUserContactMethod(UserInterface $user): void;

}
