<?php

declare(strict_types=1);

namespace Drupal\dwarl\WebAuthn;

use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Utility\Error;
use Drupal\dwarl\DwarlUserInterface;
use Drupal\dwarl\Exception\UserSaveException;
use Drupal\dwarl\Store\AnonymousStoreInterface;
use Drupal\dwarl\UserServiceInterface;
use Drupal\public_key_credential_source\Services\PublicKeyCredentialSourceWebauthn;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Register device.
 */
class RegisterDevice extends WebAuthnBase implements RegisterDeviceInterface {

  /**
   * The webauthn Framework private temp store.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStore
   */
  protected PrivateTempStore $webauthnFrameworkTempStore;

  public function __construct(
    AccountProxyInterface $currentUser,
    PublicKeyCredentialSourceWebauthn $webauthn,
    LoggerInterface $logger,
    AnonymousStoreInterface $store,
    UserServiceInterface $userService,
    RequestStack $requestStack,
    PrivateTempStoreFactory $privateTempStoreFactory,
  ) {
    parent::__construct($currentUser, $webauthn, $logger, $store, $userService, $requestStack);
    $this->webauthnFrameworkTempStore = $privateTempStoreFactory->get('webauthn_framework');
  }

  /**
   * {@inheritdoc}
   */
  public function register(string $attestation): UserInterface {

    $user = $this->store->getUser();

    // In getRegistrationOptions() the public key credential creation options
    // are saved in the private temp store.
    // At that time the user was not logged in so the session id was used in the
    // key.
    // When registerDevice() is called below the user will be logged in, so the
    // private temp store key will be different, so the public key credential
    // creation options will not be available.
    // To fix grab them now before the user is logged in.
    /* @see \Drupal\public_key_credential_source\Services\PublicKeyCredentialSourceWebauthn::getRegistrationOptions() */
    $credential_creation_options = $this->webauthnFrameworkTempStore->get('credential_creation_options');

    // The user has to be enabled for the next step to work.
    try {
      $user->activate()->save();
    }
    catch (\Throwable $exception) {
      Error::logException($this->logger, $exception);
      throw new UserSaveException($exception->getMessage(), $exception);
    }

    user_login_finalize($user);

    // And then set them again now the user is logged in.
    $this->webauthnFrameworkTempStore->set('credential_creation_options', $credential_creation_options);

    $context = ['@uid' => $user->id()];
    $this->logger->debug('Attempting to register a passkey for user @uid', $context);

    $this->webauthn->registerDevice($user, $attestation);

    return $user;
  }

  /**
   * {@inheritdoc}
   */
  public function addPasskey(string $attestation): UserInterface {
    $user = User::load(\Drupal::currentUser()->id());
    $context = ['@uid' => $user->id()];
    $this->logger->debug('Attempting to add a passkey for user @uid', $context);
    $this->webauthn->registerDevice($user, $attestation);
    return $user;
  }

}
