<?php

declare(strict_types=1);

namespace Drupal\dwarl_test\WebAuthn;

use Drupal\dwarl_test\Utils;
use Drupal\public_key_credential_source\Entity\PublicKeyCredentialSource;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Drupal\dwarl\WebAuthn\RegisterDevice as DwarlRegisterDevice;

/**
 * Register device.
 */
final class RegisterDevice extends DwarlRegisterDevice {

  public const TEST_ATTESTATION_PAYLOAD = '=== test attestation ===';

  /**
   * {@inheritdoc}
   */
  public function register(string $attestation): UserInterface {

    $user = $this->store->getUser();

    if (Utils::isValidTestUserHandle($user->getEmail()) === FALSE) {
      throw new \RuntimeException('RegisterDevice Invalid test user handle: ' . $user->getEmail());
    }

    // In getRegistrationOptions() the public key credential creation options
    // are saved in the private temp store.
    // At that time the user was not logged in so the session id was used in the
    // key.
    // When registerDevice() is called below the user will be logged in, so the
    // private temp store key will be different, so the public key credential
    // creation options will not be available.
    // To fix grab them now before the user is logged in.
    /* @see \Drupal\public_key_credential_source\Services\PublicKeyCredentialSourceWebauthn::getRegistrationOptions() */
    $credential_creation_options = $this->webauthnFrameworkTempStore->get('credential_creation_options');

    // The user has to be enabled for the next step to work.
    $user->activate()->save();

    user_login_finalize($user);

    // And then set them again now the user is logged in.
    $this->webauthnFrameworkTempStore->set('credential_creation_options', $credential_creation_options);

    $context = ['@uid' => $user->id()];
    $this->logger->debug('Attempting to register a passkey for user @uid', $context);

    if ($attestation !== self::TEST_ATTESTATION_PAYLOAD) {
      throw new \RuntimeException('Invalid test attestation');
    }

    self::createPasskey($user);

    return $user;
  }

  /**
   * {@inheritdoc}
   */
  public function addPasskey(string $attestation): UserInterface {
    $user = User::load(\Drupal::currentUser()->id());
    $context = ['@uid' => $user->id()];
    $this->logger->debug('Attempting to add a passkey for user @uid', $context);
    self::createPasskey($user);
    return $user;
  }

  /**
   * @param \Drupal\user\UserInterface $user
   *
   * @return void
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function createPasskey(UserInterface $user): void {
    $uuid = \Drupal::service('uuid');
    $values = [
      'uuid' => $uuid->generate(),
      'uid' => $user->id(),
      'created' => \Drupal::time()->getCurrentTime(),
      'public_key_credential_source' => json_encode([
        'publicKeyCredentialId' => base64_encode('test'),
        'credentialPublicKey' => base64_encode('test'),
        'userHandle' => base64_encode($user->getEmail()),
        'type' => 'test',
        'transports' => 'test',
        'attestationType' => 'test',
        'trustPath' => 'test',
        'aaguid' => 'test',
        'counter' => 'test',
      ]),
    ];
    $publicKeyCredentialSource = PublicKeyCredentialSource::create($values);
    $publicKeyCredentialSource->save();
  }

}
