<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Functional;

use Drupal\dwarl\ContactMethod;
use Drupal\Tests\dwarl\Traits\LostPasskeyTrait;
use Drupal\Tests\dwarl\Traits\WebAuthnTrait;

/**
 * Lost passkey tests.
 */
class LostPasskeyTest extends TestBase {
  use WebAuthnTrait;
  use LostPasskeyTrait;

  /**
   * Test using email with no password.
   */
  public function testLostPasskeyWithEmailNoPassword() {

    // Create user with 1 passkey and no password.
    $email = 'email@dwarl-example.com';
    $user = $this->dwarlCreateUser($email, ['validated_email'], NULL, 1);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(1, $passkeys, 'User has 1 passkey at start of test');

    // Create lost passkey request.
    $sessionId = $this->lostPasskey($email);

    // Get OTP.
    $otp = $this->getOtp($email, ContactMethod::Email);

    // Create request with valid otp.
    $data = $this->lostPasskeySubmitOtp($otp, $sessionId);
    $this->assertArrayHasKey('token', $data, 'lost_passkey.submit_otp response has token');
    $token = $data['token'];

    // Validate the token returned works.
    // This can then be used to set a new auth method.
    $this->getUserInfo($token);

    // Users passkey should have been deleted at this point.
    $user = $this->userService()->loadByUserHandle($email);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');
  }

  /**
   * Test using email with password.
   */
  public function testLostPasskeyWithEmailWithPassword() {
    // Create user with 1 passkey and password.
    $email = 'email@dwarl-example.com';
    $password = 'lEtmEiN';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password, 1);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(1, $passkeys, 'User has 1 passkey at start of test');

    // Create lost passkey request.
    $sessionId = $this->lostPasskey($email);

    // Get OTP.
    $otp = $this->getOtp($email, ContactMethod::Email);

    // Create request with valid otp.
    $data = $this->lostPasskeySubmitOtp($otp, $sessionId);
    $this->assertSame('supply password', $data['status']);

    $token = $this->lostPasskeySubmitPassword($password, $sessionId);

    // Validate the token returned works.
    // This can then be used to set a new auth method.
    $this->getUserInfo($token);

    // Users passkey should have been deleted at this point.
    $user = $this->userService()->loadByUserHandle($email);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');
  }

  /**
   * Test using email with multiple passkeys.
   */
  public function testLostPasskeyWithEmailWithMultiplePasskeys() {
    // Create user with 2 passkeys and password.
    $email = 'email@dwarl-example.com';
    $password = 'lEtmEiN';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password, 2);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(2, $passkeys, 'User has 2 passkeys at start of test');

    $sessionId = $this->lostPasskey($email);
    $otp = $this->getOtp($email, ContactMethod::Email);
    $error = $this->lostPasskeySubmitOtp($otp, $sessionId, FALSE);
    $this->assertSame(
      'Account has multiple passkeys. Use a different passkey to sign in.',
      $error, 'multiple passkeys error');
  }

  /**
   * Test using phone with no password.
   */
  public function testLostPasskeyWithPhoneNoPassword() {
    // Create user with 1 passkey and no password.
    $phone = $this->randomPhone();
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], NULL, 1);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(1, $passkeys, 'User has 1 passkey at start of test');

    // Create lost passkey request.
    $sessionId = $this->lostPasskey($phone);

    // Get OTP.
    $otp = $this->getOtp($phone, ContactMethod::Phone);

    // Create request with valid otp.
    $data = $this->lostPasskeySubmitOtp($otp, $sessionId);
    $this->assertArrayHasKey('token', $data, 'lost_passkey.submit_otp response has token');
    $token = $data['token'];

    // Validate the token returned works.
    // This can then be used to set a new auth method.
    $this->getUserInfo($token);

    // Users passkey should have been deleted at this point.
    $user = $this->userService()->loadByUserHandle($phone);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');
  }

  /**
   * Test using phone with password.
   */
  public function testLostPasskeyWithPhoneWithPassword() {
    // Create user with 1 passkey and password.
    $phone = $this->randomPhone();
    $password = 'lEtmEiN';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 1);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(1, $passkeys, 'User has 1 passkey at start of test');

    // Create lost passkey request.
    $sessionId = $this->lostPasskey($phone);

    // Get OTP.
    $otp = $this->getOtp($phone, ContactMethod::Phone);

    // Create request with valid otp.
    $data = $this->lostPasskeySubmitOtp($otp, $sessionId);
    $this->assertSame('supply password', $data['status']);

    $token = $this->lostPasskeySubmitPassword($password, $sessionId);

    // Validate the token returned works.
    // This can then be used to set a new auth method.
    $this->getUserInfo($token);

    // Users passkey should have been deleted at this point.
    $user = $this->userService()->loadByUserHandle($phone);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');
  }

  /**
   * Test using phone with multiple passkeys.
   */
  public function testLostPasskeyWithPhoneWithMultiplePasskeys() {
    // Create user with 2 passkeys and password.
    $phone = $this->randomPhone();
    $password = 'lEtmEiN';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 2);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertCount(2, $passkeys, 'User has 2 passkeys at start of test');

    $sessionId = $this->lostPasskey($phone);
    $otp = $this->getOtp($phone, ContactMethod::Phone);
    $error = $this->lostPasskeySubmitOtp($otp, $sessionId, FALSE);
    $this->assertSame(
      'Account has multiple passkeys. Use a different passkey to sign in.',
      $error, 'multiple passkeys error');
  }

}
