<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Functional;

use Drupal\dwarl\ContactMethod;
use Drupal\dwarl_test\WebAuthn\RequestOptions;
use Drupal\Tests\dwarl\Traits\ManageContactMethodsTrait;
use Drupal\user\Entity\User;

/**
 * Test setting email and phone using password and passkey.
 *
 * Also test setting preferred contact method.
 */
class ManageContactMethodsTest extends TestBase {
  use ManageContactMethodsTrait;

  /**
   * Test error when no password or passkey.
   */
  public function testSetEmailWithNoPasswordNoPasskey(): void {
    $email = 'email@dwarl-example.com';
    $user = $this->dwarlCreateUser($email, ['validated_email']);
    $token = $this->getJwtToken($user);
    $data = $this->setContactMethod($token);
    $this->assertSame('set auth method first', $data['error'], 'set_contact_method error when no password or passkey');
  }

  /**
   * Set email with password.
   */
  public function testSetEmailWithPassword(): void {
    // Create user with a phone.
    $phone = $this->randomPhone();
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password);
    $token = $this->getJwtToken($user);

    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertTrue($data['passwordRequired'], 'set_contact_method expect password required');

    // Provide password and trigger OTP send event.
    $email = 'email@dwarl-example.com';
    $this->setContactMethodRequestOtp($token, $email, $password);

    // Confirm setting new email by providing otp.
    $otp = $this->getOtp($email, ContactMethod::Email);
    $this->setContactMethodConfirm($otp, $token);

    $user = User::load($user->id());
    $this->assertSame($email, $user->getEmail(), 'User email has been updated');
    $this->assertTrue($user->hasRole('validated_email'), 'User has validated email role');
  }

  /**
   * Set email for user with passkey.
   */
  public function testSetEmailWithWithPasskey(): void {
    // Create user with a phone.
    $phone = $this->randomPhone();
    $password = NULL;
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 1);
    $token = $this->getJwtToken($user);

    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertSame('=== test authentication request ===', $data['requestOptions']['challenge'], 'set_contact_method expect requestOptions');

    // Provide password and trigger OTP send event.
    $email = 'test-email@dwarl-example.com';
    $this->setContactMethodRequestOtp($token, $email, NULL, RequestOptions::TEST_CHALLENGE);

    // Confirm setting new email by providing otp.
    $otp = $this->getOtp($email, ContactMethod::Email);
    $this->setContactMethodConfirm($otp, $token);

    $user = User::load($user->id());
    $this->assertSame($email, $user->getEmail(), 'User email has been updated');
  }

  /**
   * Set phone with no password or passkey.
   */
  public function testSetPhoneWithNoPasswordNoPasskey(): void {
    $phone = $this->randomPhone();
    $user = $this->dwarlCreateUser($phone, ['validated_email']);
    $token = $this->getJwtToken($user);
    $data = $this->setContactMethod($token);
    $this->assertSame('set auth method first', $data['error'], 'set_contact_method error when no password or passkey');
  }

  /**
   * Set phone with password.
   */
  public function testSetPhoneWithPassword(): void {
    // Create user with a email.
    $email = 'test@dwarl-example.com';
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password);
    $token = $this->getJwtToken($user);
    $phone = $this->setPhone($token, $password);

    $user = User::load($user->id());
    $userPhone = $user->get('field_phone')->getValue();
    $this->assertSame($phone, $userPhone[0]['value'], 'Phone has been set');
    $this->assertTrue($user->hasRole('validated_phone'), 'User has validated phone role');
  }

  /**
   * Set phone with passkey.
   */
  public function testSetPhoneWithPasskey(): void {
    // Create user with a email.
    $email = 'test@dwarl-example.com';
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password, 1);
    $token = $this->getJwtToken($user);

    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertSame('=== test authentication request ===', $data['requestOptions']['challenge'], 'set_contact_method expect requestOptions');

    // Provide password and trigger OTP send event.
    $phone = $this->randomPhone();
    $this->setContactMethodRequestOtp($token, $phone, NULL, RequestOptions::TEST_CHALLENGE);

    // Confirm setting new phone by providing otp.
    $otp = $this->getOtp($phone, ContactMethod::Phone);
    $this->setContactMethodConfirm($otp, $token);

    $user = User::load($user->id());
    $userPhone = $user->get('field_phone')->getValue();
    $this->assertSame($phone, $userPhone[0]['value'], 'Phone has been set');
    $this->assertTrue($user->hasRole('validated_phone'), 'User has validated phone role');
  }

  /**
   * Update email with password.
   *
   * See https://cheatsheetseries.owasp.org/cheatsheets/Authentication_Cheat_Sheet.html#changing-a-users-registered-email-address.
   */
  public function testUpdateEmailWithPassword(): void {
    // Create user with email.
    $email = 'test1@dwarl-example.com';
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password);
    $token = $this->getJwtToken($user);

    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertTrue($data['passwordRequired'], 'set_contact_method expect password required');

    // Provide password and trigger OTP send event.
    $updatedEmail = 'test2@dwarl-example.com';
    $this->setContactMethodRequestOtp($token, $updatedEmail, $password);

    // Confirm setting new email by providing otp.
    $otp = $this->getOtp($updatedEmail, ContactMethod::Email);
    $this->setContactMethodConfirm($otp, $token);

    $user = User::load($user->id());
    $this->assertSame($updatedEmail, $user->getEmail(), 'User email has been updated');
    $this->assertTrue($user->hasRole('validated_email'), 'User has validated email role');
  }

  /**
   * Update phone with passkey.
   */
  public function testUpdatePhoneWithPasskey(): void {
    // Create user with a email.
    $phone = $this->randomPhone();
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 1);
    $token = $this->getJwtToken($user);

    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertSame('=== test authentication request ===', $data['requestOptions']['challenge'], 'set_contact_method expect requestOptions');

    // Provide auth request and trigger OTP send event.
    $updatedPhone = $this->randomPhone();
    $this->setContactMethodRequestOtp($token, $updatedPhone, NULL, RequestOptions::TEST_CHALLENGE);

    // Confirm setting new phone by providing otp.
    $otp = $this->getOtp($updatedPhone, ContactMethod::Phone);
    $this->setContactMethodConfirm($otp, $token);

    $user = User::load($user->id());
    $userPhone = $user->get('field_phone')->getValue();
    $this->assertSame($updatedPhone, $userPhone[0]['value'], 'Phone has been set');
    $this->assertTrue($user->hasRole('validated_phone'), 'User has validated phone role');
  }

  /**
   * Test set preferred contact method.
   */
  public function testSetPreferredContactMethod(): void {
    // Create user with a email.
    $email = 'something@test.com';
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($email, ['validated_phone'], $password);
    $token = $this->getJwtToken($user);

    $pcm = $user->get('field_preferred_contact_method')->getString();
    $this->assertSame('email', $pcm, 'PCM is email at start of test');

    // Try and set preferred contact method to phone.
    $response = $this->setPreferredContractMethod('phone', $token);
    $this->assertSame(422, $response->getStatusCode(), 'set_preferred_contact_method to phone with no phone');

    // Set a phone.
    // Make a call to set contact method.
    $this->setPhone($token, $password);

    // Check PCM is still email after phone has been set.
    $user = User::load($user->id());
    $pcm = $user->get('field_preferred_contact_method')->getString();
    $this->assertSame('email', $pcm, 'PCM is still email after phone set');

    // Set preferred contact method to phone.
    $response = $this->setPreferredContractMethod('phone', $token);
    $this->assertSame(200, $response->getStatusCode(), 'set_preferred_contact_method to phone');

    // Check PCM is now phone.
    \Drupal::entityTypeManager()->getStorage('user')->resetCache([$user->id()]);
    $user = User::load($user->id());
    $pcm = $user->get('field_preferred_contact_method')->getString();
    $this->assertSame('phone', $pcm, 'PCM is now phone');
  }

  /**
   * @param string $token
   * @param string $password
   *
   * @return string
   */
  private function setPhone(string $token, string $password): string {
    // Make a call to set contact method.
    $data = $this->setContactMethod($token);
    $this->assertTrue($data['passwordRequired'], 'set_contact_method expect password required');

    // Provide password and trigger OTP send event.
    $phone = $this->randomPhone();
    $this->setContactMethodRequestOtp($token, $phone, $password);

    // Confirm setting new email by providing otp.
    $otp = $this->getOtp($phone, ContactMethod::Phone);
    $this->setContactMethodConfirm($otp, $token);
    return $phone;
  }

}
