<?php

namespace Drupal\Tests\dwarl\Functional;

use Drupal\dwarl\ContactMethod;
use Drupal\Tests\dwarl\Traits\ManagePasswordsTrait;
use Drupal\user\Entity\User;

/**
 * Manage password tests.
 */
class ManagePasswordsTest extends TestBase {

  use ManagePasswordsTrait;

  /**
   * Set password by sending an OTP via email.
   */
  public function testSetPasswordWithEmail(): void {
    // Create user with passkey but no password.
    $email = 'test@dwarl-example.com';
    $user = $this->dwarlCreateUser($email, ['validated_email'], NULL, 1);
    $token = $this->getJwtToken($user);

    // Make request to trigger sending of OTP via email.
    $this->requestSetPasswordOtp('email', $token);

    // Provide OTP and set password.
    $otp = $this->getOtp($email, ContactMethod::Email);
    $password = 'changeMe';
    $this->setPasswordWithOtpRequest($password, $otp, $token);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($password, $user->getPassword()), 'password is valid');
  }

  /**
   * Set password by sending an OTP via phone.
   */
  public function testSetPasswordWithPhone(): void {
    // Create user with passkey but no password.
    $phone = $this->randomPhone();
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], NULL, 1);
    $token = $this->getJwtToken($user);

    // Make request to trigger sending of OTP via email.
    $this->requestSetPasswordOtp('phone', $token);

    // Provide OTP and set password.
    $otp = $this->getOtp($phone, ContactMethod::Phone);
    $password = 'changeMe';
    $this->setPasswordWithOtpRequest($password, $otp, $token);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($password, $user->getPassword()), 'password is valid');
  }

  /**
   * Set password with passkey.
   */
  public function testSetPasswordWithPasskey(): void {
    // Create user with passkey but no password.
    $phone = $this->randomPhone();
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], NULL, 1);
    $token = $this->getJwtToken($user);

    // Set password with passkey.
    $password = 'SuperSecurePassword';
    $this->setPasswordWithPasskey($token, $password);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($password, $user->getPassword()), 'password is valid');
  }

  /**
   * Update password by sending OTP via email.
   */
  public function testChangePasswordWithEmail(): void {
    // Create user with passkey and password.
    $email = 'test@dwarl-example.com';
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($email, ['validated_email'], $password, 1);
    $token = $this->getJwtToken($user);

    // Expect fail from set password endpoint.
    $this->requestSetPasswordOtp('email', $token, 422);

    // Call update password endpoint.
    $this->requestChangePasswordOtp('email', $password, $token);

    // Provide OTP and set password.
    $otp = $this->getOtp($email, ContactMethod::Email);
    $password = 'changeMe';
    $this->setPasswordWithOtpRequest($password, $otp, $token);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($password, $user->getPassword()), 'password is valid');
  }

  /**
   * Update password by sending OTP via phone.
   */
  public function testChangePasswordWithPhone(): void {
    // Create user with passkey and password.
    $phone = $this->randomPhone();
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 1);
    $token = $this->getJwtToken($user);

    // Expect fail from set password endpoint.
    $this->requestSetPasswordOtp('phone', $token, 422);

    // Call update password endpoint.
    $this->requestChangePasswordOtp('phone', $password, $token);

    // Provide OTP and set password.
    $otp = $this->getOtp($phone, ContactMethod::Phone);
    $password = 'changeMe';
    $this->setPasswordWithOtpRequest($password, $otp, $token);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($password, $user->getPassword()), 'password is valid');
  }

  /**
   * Update password with passkey.
   */
  public function testChangePasswordWithPasskey(): void {
    // Create user with passkey but no password.
    $phone = $this->randomPhone();
    $password = 'changeMe';
    $user = $this->dwarlCreateUser($phone, ['validated_phone'], $password, 1);
    $token = $this->getJwtToken($user);

    // Set password with passkey.
    $updatedPassword = 'SuperSecurePassword';
    $this->setPasswordWithPasskey($token, $updatedPassword, $password);

    // Check user password has been set.
    $user = User::load($user->id());

    /** @var \Drupal\Core\Password\PhpPassword $passwordService */
    $passwordService = \Drupal::service('password');

    $this->assertTrue($passwordService->check($updatedPassword, $user->getPassword()), 'password is valid');
  }

}
