<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Functional;

use Drupal\dwarl\ContactMethod;
use Drupal\Tests\dwarl\Traits\RegisterTrait;
use Drupal\Tests\dwarl\Traits\WebAuthnTrait;
use Drupal\user\UserInterface;

/**
 * Dwarl registration tests.
 */
class RegisterTest extends TestBase {
  use RegisterTrait;
  use WebAuthnTrait;

  /**
   * Register with email and password.
   */
  public function testRegisterEmailPassword(): void {
    [$sessionId, $email] = $this->emailRegistrationFirstStep();

    // Set password.
    $password = 'GNUTerryPratchett';
    $jwtToken = $this->registerSetPassword($password, $sessionId);

    $this->assertNotEmpty($jwtToken, 'There is a JWT token');

    // User should have 0 passkeys.
    $user = $this->userService()->loadByUserHandle($email);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys);

    $userInfo = $this->getUserInfo($jwtToken);

    $this->assertSame($email, $userInfo['email'], 'User info email matches');
    $this->assertSame('', $userInfo['phone'], 'Phone is not set');
    $this->assertSame('email', $userInfo['preferredContactMethod'], 'preferredContactMethod is email');
    $this->assertTrue($userInfo['hasPassword'], 'hasPassword is true');
    $this->assertEquals(0, count($userInfo['passkeys']), 'there are 0 passkeys');
  }

  /**
   * Register with phone and password.
   */
  public function testRegisterPhonePassword(): void {
    $phone = $this->randomPhone();

    $user = $this->userService()->loadByUserHandle($phone);
    $this->assertNull($user, "user $phone does not exist at start of test");

    $sessionId = $this->phoneRegistrationFirstStep($phone);

    // Set password.
    $password = 'electricBoogaloo2';
    $jwtToken = $this->registerSetPassword($password, $sessionId);

    // User should have 0 passkeys.
    $user = $this->userService()->loadByUserHandle($phone);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys);

    // $this->assertTrue($user->hasField('field_preferred_contact_method'), 'User has field_preferred_contact_method');

    $userInfo = $this->getUserInfo($jwtToken);

    $this->assertSame($phone, $userInfo['phone'], 'User info phone matches');
    $this->assertSame('', $userInfo['email'], 'Email is not set');
    // $this->assertSame('phone', $userInfo['preferredContactMethod'], 'preferredContactMethod is phone');
    $this->assertTrue($userInfo['hasPassword'], 'hasPassword is true');
    $this->assertEquals(0, count($userInfo['passkeys']), 'there are 0 passkeys');

  }

  /**
   * Register with email and passkey.
   */
  public function testRegisterEmailPasskey(): void {
    [$sessionId, $email] = $this->emailRegistrationFirstStep();

    // Create passkey.
    $userInfo = $this->createPasskey($sessionId, $email);

    $this->assertSame($email, $userInfo['email'], 'User info email matches');
    $this->assertSame('', $userInfo['phone'], 'Phone is not set');
    $this->assertSame('email', $userInfo['preferredContactMethod'], 'preferredContactMethod is email');
    $this->assertFalse($userInfo['hasPassword'], 'hasPassword is false');
    $this->assertEquals(1, count($userInfo['passkeys']), 'there is one passkey');
    $passkey = reset($userInfo['passkeys']);
    $this->assertPasskeyIsValid($passkey);
  }

  /**
   * Register with phone and passkey.
   */
  public function testRegisterPhonePasskey(): void {
    $phone = $this->randomPhone();

    $user = $this->userService()->loadByUserHandle($phone);
    $this->assertNull($user, "user $phone does not exist at start of test");

    $sessionId = $this->phoneRegistrationFirstStep($phone);

    // Create passkey.
    $userInfo = $this->createPasskey($sessionId, $phone);

    $this->assertSame($phone, $userInfo['phone'], 'User info phone matches');
    $this->assertSame('',$userInfo['email'], 'Email is not set');
    // $this->assertSame('phone', $userInfo['preferredContactMethod'], 'preferredContactMethod is phone');
    $this->assertFalse($userInfo['hasPassword'], 'hasPassword is false');
    $this->assertEquals(1, count($userInfo['passkeys']), 'there is one passkey');
    $passkey = reset($userInfo['passkeys']);
    $this->assertPasskeyIsValid($passkey);
  }

  /**
   * Try to register with an existing phone number.
   * @broken
   */
  public function testRegisterWithExistingPhone(): void {
    $phone = '+447758930294';
    $this->dwarlCreateUser($phone, ['validated_phone'], 'password');

    // Create user.
    $this->createUserPhone($phone, TRUE);

    // Ensure no OTP is available.
    $expectedEmail = "$phone@dwarl.com";
    $this->getOtp($expectedEmail, ContactMethod::Phone, TRUE);
  }

  /**
   * Try to register with an existing email.
   * @broken
   */
  public function testRegisterWithExistingEmail(): void {
    $email = 'emailOne@dwarl-example.com';
    $this->dwarlCreateUser($email, ['validated_email'], 'password');

    $this->createUserEmail($email, TRUE);

    // Ensure no OTP is available.
    $this->getOtp($email, ContactMethod::Email, TRUE);
  }

  /**
   * First step of email registration.
   *
   * @return array
   *   Session id and email.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function emailRegistrationFirstStep(): array {
    $email = $this->randomMachineName() . '@dwarl-example.com';

    $user = user_load_by_mail($email);
    $this->assertFalse($user, "user $email does not exist at start of test");

    // Create user.
    $sessionId = $this->createUserEmail($email, TRUE);

    // Get users otp code.
    $otp = $this->getOtp($email, ContactMethod::Email);

    // Validate email.
    $user = $this->validateContactMethod($email, $sessionId, $otp);
    $this->assertTrue($user->hasRole('validated_email'), 'user has validated email role');

    $this->assertPreferredContactMethod($user, 'email');

    return [$sessionId, $email];
  }

  /**
   * Phone registration first step.
   *
   * @return string
   *   Session id.
   */
  private function phoneRegistrationFirstStep(string $phone): string {

    // Create user.
    $sessionId = $this->createUserPhone($phone, TRUE);

    // Get users otp code.
    $otp = $this->getOtp($phone, ContactMethod::Phone);

    // Validate phone.
    $user = $this->validateContactMethod($phone, $sessionId, $otp);

    $this->assertTrue($user->hasRole('validated_phone'), 'user has validated phone role');

    $this->assertPreferredContactMethod($user, 'phone');

    return $sessionId;
  }

  /**
   * Assert preferred contact method.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   * @param string $expectedContactMethod
   *   Expected contact method.
   */
  public function assertPreferredContactMethod(UserInterface $user, string $expectedContactMethod): void {
    $hasField = $user->hasField('field_preferred_contact_method');
    $this->assertTrue($hasField, 'Has field_preferred_contact_method');

    $pcm = $user->get('field_preferred_contact_method')->getString();
    $this->assertSame($expectedContactMethod, $pcm, 'Preferred contact method is ' . $expectedContactMethod);
  }

  /**
   * Create passkey.
   *
   * @param string $sessionId
   *   The session id.
   * @param string $userHandle
   *   THe user handle.
   *
   * @return array
   */
  private function createPasskey(string $sessionId, string $userHandle): array {
    $optionsJson = $this->registerDeviceOptions($sessionId, 'dwarl.registration.register_device_options');
    $jwtToken = $this->registerDevice($sessionId, $userHandle, $optionsJson, 'dwarl.registration.register_device');

    $this->assertNotEmpty($jwtToken, 'There is a JWT token');

    // User should have 1 passkey.
    $user = $this->userService()->loadByUserHandle($userHandle);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertPasskeyIsValid($passkeys[0]);

    return $this->getUserInfo($jwtToken);
  }

}
