<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Functional;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\dwarl\ContactMethod;
use Drupal\dwarl\DwarlUserInterface;
use Drupal\dwarl\PasskeyStorageInterface;
use Drupal\dwarl\UserService;
use Drupal\dwarl_test\WebAuthn\RegisterDevice;
use Drupal\jwt\Authentication\Provider\JwtAuth;
use Drupal\Tests\ApiRequestTrait;
use Drupal\Tests\BrowserTestBase;
use Drupal\user\Entity\Role;
use Drupal\user\Entity\User;
use Drupal\user\RoleInterface;
use Drupal\user\UserInterface;
use GuzzleHttp\RequestOptions;

/**
 * Dwarl test base.
 */
class TestBase extends BrowserTestBase {
  use ApiRequestTrait;

  private const OTP_CODE_LENGTH = 6;

  /**
   * Modules installed for all tests.
   *
   * @var array
   */
  protected static $modules = [
    'system',
    'user',
    'router_test',
    'key',
    'jwt',
    'jwt_test',
    'jwt_auth_issuer',
    'jwt_auth_consumer',
    'public_key_credential_source',
    'webauthn_framework',
    'dwarl',
    'dwarl_test',
    'dwarl_phone',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Passkey storage.
   *
   * @var \Drupal\dwarl\PasskeyStorageInterface|mixed
   */
  protected PasskeyStorageInterface $passkeyStorage;

  /**
   * Dwarl user.
   *
   * @var \Drupal\dwarl\DwarlUserInterface|mixed
   */
  protected DwarlUserInterface $userService;

  /**
   * Get Dwarl user service.
   *
   * @return \Drupal\dwarl\UserService
   *   The Dwarl user service.
   */
  protected function userService(): UserService {
    if (!isset($this->userService)) {
      $this->userService = \Drupal::service('dwarl.user_service');
    }
    return $this->userService;
  }

  /**
   * Get passkey storage.
   *
   * @return \Drupal\dwarl\PasskeyStorageInterface
   *   Passkey storage.
   */
  protected function passkeyStorage(): PasskeyStorageInterface {
    if (!isset($this->passkeyStorage)) {
      $this->passkeyStorage = \Drupal::service('dwarl.passkey_storage');
    }
    return $this->passkeyStorage;

  }

  /**
   * Generate random phone number.
   *
   * @return string
   *   The phone number.
   */
  public function randomPhone(): string {
    return '+447761730' . rand(0, 9) . rand(0, 9) . rand(0, 9);
  }

  /**
   * Create a test user.
   *
   * @param string $userHandle
   *   Email or phone.
   * @param array $roles
   *   Roles.
   * @param string|null $password
   *   Password.
   * @param int $passkeyCount
   *   Passkey count.
   * @param string|null $preferredContactMethod
   *   Preferred contact method.
   *
   * @return \Drupal\user\UserInterface
   *   The new user.
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function dwarlCreateUser(
    string $userHandle,
    array $roles = [],
    ?string $password = NULL,
    int $passkeyCount = 0,
    ?string $preferredContactMethod = NULL
  ): UserInterface {

    $email = str_contains($userHandle, '@') ? $userHandle : NULL;
    $phone = $email ? NULL : $userHandle;

    $values = [
      'status' => 1,
      'field_preferred_contact_method' => $preferredContactMethod,
    ];

    if ($email) {
      $values['mail'] = $email;
      $values['name'] = $email;
      if ($preferredContactMethod === NULL) {
        $values['field_preferred_contact_method']['value'] = 'email';
      }
    }
    elseif ($phone) {
      $values['field_phone'] = [
        'value' => $phone,
        'country' => 'GB',
      ];
      $values['mail'] = "$phone@dwarl.com";
      $values['name'] = $phone;
      if ($preferredContactMethod === NULL) {
        $values['field_preferred_contact_method']['value'] = 'phone';
      }
    }
    else {
      throw new \RuntimeException('Email or phone number is required.');
    }

    $user = User::create($values);

    if ($password) {
      $user->setPassword($password);
    }

    if ($roles) {
      foreach ($roles as $role) {
        $user->addRole($role);
      }
    }

    $user->save();

    $valid_user = $user->id() !== NULL;
    $this->assertTrue($valid_user, "User created with name {$values['name']}");

    for ($i = 0; $i < $passkeyCount; $i++) {
      RegisterDevice::createPasskey($user);
    }

    return $user;
  }

  /**
   * Get JWT token.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user.
   *
   * @return string
   *   The JWT token.
   */
  protected function getJwtToken(UserInterface $user): string {
    /** @var JwtAuth $jwtAuth */
    $jwtAuth = \Drupal::service('jwt.authentication.jwt');

    user_login_finalize($user);
    return $jwtAuth->generateToken();
  }

  /**
   * Get OTP code.
   *
   * @param string $expectedUserHandle
   *   The expected email.
   * @param \Drupal\dwarl\ContactMethod $expectedMethod
   *   The expected OTP send method.
   * @param bool $expectNull
   *   Should we expect the OTP to not be available?
   *
   * @return string|null
   *   The otp or null.
   */
  protected function getOtp(string $expectedUserHandle, ContactMethod $expectedMethod, bool $expectNull = FALSE): ?string {

    \Drupal::request()->getSession()->set('core.tempstore.private.owner', 'dwarl_test');

    /** @var \Drupal\Core\TempStore\PrivateTempStoreFactory $tempStoreFactory */
    $tempStoreFactory = \Drupal::service('tempstore.private.one_hour');
    $tempStore = $tempStoreFactory->get('dwarl_test');

    $otp = $tempStore->get('otp');
    $email = $tempStore->get('email');
    $phone = $tempStore->get('phone');
    $method = $tempStore->get('method');

    if ($expectNull) {
      $this->assertEmpty($otp, 'otp is empty');
      return NULL;
    }

    // Check OTP length.
    $this->assertEquals(self::OTP_CODE_LENGTH, strlen($otp), 'otp length check');
    $this->assertEquals($expectedMethod, $method, 'method match');

    if ($method === ContactMethod::Phone) {
      $this->assertEquals($expectedUserHandle, $phone, 'phone OTP match');
    }
    else {
      $this->assertEquals($expectedUserHandle, $email, 'email OTP match');
    }

    return $otp;
  }

  /**
   * Get user info.
   *
   * @param string $token
   *   The JWT token.
   *
   * @return array
   *   The user info.
   */
  protected function getUserInfo(string $token): array {
    $url = Url::fromRoute('dwarl.user_info');

    // Make request with wrong token.
    $request_options = [
      RequestOptions::HEADERS => [
        'Authorization' => "Bearer nope",
        'Access-Control-Allow-Headers' => 'Authorization',
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'user_info without token');

    $request_options[RequestOptions::HEADERS]['Authorization'] = "Bearer $token";
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'user_info response code');
    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('email', $data);
    $this->assertArrayHasKey('phone', $data);
    $this->assertArrayHasKey('preferredContactMethod', $data);
    $this->assertArrayHasKey('hasPassword', $data);
    $this->assertArrayHasKey('passkeys', $data);

    return $data;
  }

}
