<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use GuzzleHttp\RequestOptions;

/**
 * Methods for account recovery tests.
 */
trait AccountRecoveryTrait {

  /**
   * Request options.
   *
   * @param string $token
   *   The JWT token.
   *
   * @return string
   *   The auth request.
   */
  protected function requestOptions(string $token): string {
    $url = Url::fromRoute('dwarl.recovery.request_options');

    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'recovery.request_options response');

    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('challenge', $data['requestOptions'], 'recovery.request_options has challenge');

    return $data['requestOptions']['challenge'];
  }

  /**
   * Generate account recovery code.
   *
   * @param string $authRequest
   *   The auth request.
   * @param string $token
   *   The token.
   *
   * @return string
   *   The recovery code.
   */
  protected function generateRecoveryCode(string $authRequest, string $token): string {
    $url = Url::fromRoute('dwarl.recovery.generate_recovery_code');
    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
      RequestOptions::FORM_PARAMS => ['authentication_request' => $authRequest],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'recovery.request_options response');

    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('recovery_code', $data, 'has recovery_code');

    return $data['recovery_code'];
  }

  /**
   * Recover account.
   *
   * @param string $userHandle
   *   The user handle.
   * @param string $recoveryCode
   *   The recovery code.
   *
   * @return string
   *   A JWT token.
   */
  protected function recoverAccount(string $userHandle, string $recoveryCode): string {
    $url = Url::fromRoute('dwarl.recovery.recover_account');
    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'user_handle' => $userHandle,
        'recovery_code' => 'wrongCode',
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(422, $response->getStatusCode(), 'recovery.request_options with wrong code');

    $request_options[RequestOptions::FORM_PARAMS]['recovery_code'] = $recoveryCode;

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'recovery.request_options response');

    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('token', $data, 'has token');

    return $data['token'];
  }

}
