<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use GuzzleHttp\RequestOptions;

/**
 * Methods for forgot password tests.
 */
trait ForgotPasswordTrait {

  /**
   * Call forgot password endpoint.
   *
   * @param string $userHandle
   *   User handle.
   * @param bool $assignPermission
   *   Should permissions be assigned?
   *
   * @return string
   *   The session id.
   */
  protected function forgotPassword(string $userHandle, bool $assignPermission = TRUE): string {
    $url = Url::fromRoute('dwarl.forgot_password');
    $request_options = [
      RequestOptions::FORM_PARAMS => ['user_handle' => $userHandle],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);

    if ($assignPermission) {
      $this->assertSame(403, $response->getStatusCode(), 'forgot_password response without permission');

      // Now assign the "dwarl change own password" permission.
      $anon_role = Role::load(RoleInterface::ANONYMOUS_ID);
      $this->grantPermissions($anon_role, ['dwarl change own password']);
    }

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'forgot_password response');

    $headers = $response->getHeaders();
    $this->assertArrayHasKey('Session-Id', $headers);

    return $headers['Session-Id'][0];
  }

  /**
   * Forgot password set password.
   *
   * @param string $password
   *   The password to set.
   * @param string $otp
   *   The OTP code.
   * @param string $sessionId
   *   The session id.
   *
   * @return array
   *   The response body.
   */
  protected function forgotPasswordSetPassword(string $password, string $otp, string $sessionId): array {
    $url = Url::fromRoute('dwarl.forgot_password_set_password');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'password' => $password,
        'otp' => $otp,
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'forgot_password_set_password without session');

    $request_options[RequestOptions::HEADERS]['session-id'] = $sessionId;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'forgot_password_set_password response');

    return Json::decode($response->getBody());
  }

  /**
   * Forgot password set password confirm.
   *
   * @param string $password
   *   The password.
   * @param string $authRequest
   *   The auth request.
   * @param string $sessionId
   *   The session id.
   *
   * @return string
   *   JWT token.
   */
  protected function forgotPasswordSetPasswordConfirm(string $password, string $authRequest, string $sessionId): string {
    $url = Url::fromRoute('dwarl.forgot_password_set_password_confirm');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'password' => $password,
        'authentication_request' => $authRequest,
      ],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'forgot_password_set_password_confirm without session');

    $request_options[RequestOptions::HEADERS]['session-id'] = $sessionId;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'forgot_password_set_password response');

    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('token', $data);

    return $data['token'];
  }

}
