<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use GuzzleHttp\RequestOptions;

/**
 * Methods for lost passkey tests.
 */
trait LostPasskeyTrait {

  /**
   * Lost passkey request.
   *
   * @param string $userHandle
   *   User handle.
   *
   * @return string
   *   Session ID.
   */
  protected function lostPasskey(string $userHandle): string {
    $url = Url::fromRoute('dwarl.lost_passkey');

    // Make request with email that doesn't exist.
    // Response should be the same but not OPT will be sent.
    $request_options = [
      RequestOptions::FORM_PARAMS => ['user_handle' => 'not-exist@email.com'],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'lost_passkey');
    $headers = $response->getHeaders();
    $this->assertArrayHasKey('Session-Id', $headers);

    // Make request with real email. This time an OPT should be sent.
    $request_options = [
      RequestOptions::FORM_PARAMS => ['user_handle' => $userHandle],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'lost_passkey');

    $headers = $response->getHeaders();
    $this->assertArrayHasKey('Session-Id', $headers);

    return $headers['Session-Id'][0];
  }

  /**
   * Lost passkey submit OTP.
   *
   * @param string $otp
   *   The OTP.
   * @param string $sessionId
   *   The session ID.
   * @param bool $expectSuccess
   *   Is success expected?
   *
   * @return array|string
   *   If success is expected return the response body, else the error string.
   */
  protected function lostPasskeySubmitOtp(string $otp, string $sessionId, bool $expectSuccess = TRUE): array|string {
    $url = Url::fromRoute('dwarl.lost_passkey.submit_otp');

    $request_options = [
      RequestOptions::FORM_PARAMS => ['otp' => $otp],
    ];
    // Make request without session id.
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'lost_passkey.submit_otp without session');

    $request_options[RequestOptions::HEADERS]['session-id'] = $sessionId;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $data = Json::decode($response->getBody());

    if ($expectSuccess) {
      $this->assertSame(200, $response->getStatusCode(), 'lost_passkey');
      $this->assertArrayHasKey('status', $data, 'lost_passkey.submit_otp response has status');
      return $data;
    }

    $this->assertArrayHasKey('error', $data);

    return $data['error'];
  }

  /**
   * Lost passkey submit password.
   *
   * @param string $password
   *   The password.
   * @param string $sessionId
   *   The session id.
   *
   * @return string
   *   A JWT token.
   */
  protected function lostPasskeySubmitPassword(string $password, string $sessionId): string {
    $url = Url::fromRoute('dwarl.lost_passkey.submit_password');
    $request_options = [
      RequestOptions::FORM_PARAMS => ['password' => $password],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'lost_passkey.submit_password without session');

    $request_options[RequestOptions::HEADERS]['session-id'] = $sessionId;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'lost_passkey.submit_password response');

    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('token', $data, 'lost_passkey.submit_password response has token');
    return $data['token'];
  }

}
