<?php

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use GuzzleHttp\RequestOptions;
use Psr\Http\Message\ResponseInterface;

/**
 * Methods for manage contact methods test.
 */
trait ManageContactMethodsTrait {

  /**
   * Set contact method.
   *
   * @param string $token
   *   JWT token.
   *
   * @return array
   *   The response body.
   */
  protected function setContactMethod(string $token): array {
    $url = Url::fromRoute('dwarl.set_contact_method');

    $response = $this->makeApiRequest('POST', $url, []);
    $this->assertSame(403, $response->getStatusCode(), 'set_contact_method without token');

    $request_options[RequestOptions::HEADERS]['Authorization'] = "Bearer $token";
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'set_contact_method response code');
    return Json::decode($response->getBody());
  }

  /**
   * Set contact method request OTP.
   *
   * @param string $token
   *   The JWT token.
   * @param string|null $password
   *   The password.
   * @param string|null $authRequest
   *   The auth request.
   * @param string $userHandle
   *   The user handle to set.
   */
  protected function setContactMethodRequestOtp(string $token, string $userHandle, ?string $password = NULL, ?string $authRequest = NULL): void {
    $url = Url::fromRoute('dwarl.set_contact_method_request_otp');

    // authentication_request, email

    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
      RequestOptions::FORM_PARAMS => [
        'password' => 'wrongPassword',
        'userHandle' => $userHandle,
      ],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(422, $response->getStatusCode(), 'set_contact_method_request_otp with wrong password');

    if ($password) {
      $request_options[RequestOptions::FORM_PARAMS]['password'] = $password;
    }
    if ($authRequest) {
      $request_options[RequestOptions::FORM_PARAMS]['authentication_request'] = $authRequest;
    }

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'set_contact_method_request_otp');
  }

  protected function setContactMethodConfirm(string $otp, string $token) {
    $url = Url::fromRoute('dwarl.set_contact_method_confirm');

    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
      RequestOptions::FORM_PARAMS => [
        'otp' => 'wrongOtp',
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(400, $response->getStatusCode(), 'set_contact_method_confirm with wrong otp');

    $request_options[RequestOptions::FORM_PARAMS]['otp'] = $otp;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'set_contact_method_confirm');
  }

  /**
   * Set preferred contract method.
   *
   * @param string $method
   *   The method.
   * @param string $token
   *   The token.
   *
   * @return \Psr\Http\Message\ResponseInterface
   */
  protected function setPreferredContractMethod(string $method, string $token): ResponseInterface {
    $url = Url::fromRoute('dwarl.set_preferred_contact_method');

    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
      RequestOptions::FORM_PARAMS => [
        'method' => $method,
      ],
    ];

    return $this->makeApiRequest('POST', $url, $request_options);
  }

}
