<?php

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\user\Entity\Role;
use GuzzleHttp\RequestOptions;

/**
 * Methods for manage passkeys tests.
 */
trait ManagePasskeysTrait {

  /**
   * Add passkey.
   *
   * @param string $token
   *   A JWT token.
   * @param bool $setPermissions
   *   Should permissions be set?
   */
  protected function addPasskey(string $token, bool $setPermissions): void {
    $url = Url::fromRoute('dwarl.authenticated.register_device_options');

    $response = $this->makeApiRequest('POST', $url, []);
    $this->assertSame(403, $response->getStatusCode(), 'authenticated.register_device_options without token');

    $request_options[RequestOptions::HEADERS]['Authorization'] = "Bearer $token";
    $response = $this->makeApiRequest('POST', $url, $request_options);

    if ($setPermissions) {
      $this->assertSame(403, $response->getStatusCode(), 'authenticated.register_device_options without permission');

      // Assign 'dwarl manage own passkeys' permission.
      $auth_role = Role::load(AccountInterface::AUTHENTICATED_ROLE);
      $this->grantPermissions($auth_role, ['dwarl manage own passkeys']);

      $response = $this->makeApiRequest('POST', $url, $request_options);
    }

    $this->assertSame(200, $response->getStatusCode(), 'authenticated.register_device_options response code');

    $url = Url::fromRoute('dwarl.authenticated.register_device');
    $attestation = 'test';
    $request_options = [
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
      RequestOptions::FORM_PARAMS => [
        'attestation' => $attestation,
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'authenticated.register_device response code');
  }

  /**
   * Delete passkey.
   *
   * @param string $token
   *   A JWT token.
   * @param int $id
   *   The passkey id.
   */
  protected function deletePasskey(string $token, int $id): void {
    $params = ['public_key_credential_source' => $id];
    $url = Url::fromRoute('dwarl.passkey.delete', $params);

    $response = $this->makeApiRequest('POST', $url, []);
    $this->assertSame(403, $response->getStatusCode(), 'passkey.delete without token');

    $request_options[RequestOptions::HEADERS]['Authorization'] = "Bearer $token";
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'passkey.delete without permission');

    // Assign 'dwarl manage own passkeys' permission.
    $auth_role = Role::load(AccountInterface::AUTHENTICATED_ROLE);
    $this->grantPermissions($auth_role, ['dwarl manage own passkeys']);

    $response = $this->makeApiRequest('POST', $url, $request_options);

    $this->assertSame(200, $response->getStatusCode(), 'passkey.delete');
  }

}
