<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use GuzzleHttp\RequestOptions;

/**
 * Methods for manage password tests.
 */
trait ManagePasswordsTrait {

  /**
   * Request set password OTP.
   *
   * @param string $method
   *   The OTP send method.
   * @param string $token
   *   The JWT token.
   * @param int $expectedCode
   *   The expected response code.
   */
  protected function requestSetPasswordOtp(string $method, string $token, int $expectedCode = 200): void {
    $url = Url::fromRoute('dwarl.request_set_password_otp');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'method' => $method,
      ],
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame($expectedCode, $response->getStatusCode(), 'request_change_password_otp response');
  }

  /**
   * Request change password OTP.
   *
   * @param string $method
   *   The otp send method.
   * @param string $password
   *   The current password.
   * @param string $token
   *   The JWT token.
   */
  protected function requestChangePasswordOtp(string $method, string $password, string $token): void {
    $url = Url::fromRoute('dwarl.request_change_password_otp');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'method' => $method,
        'password' => $password,
      ],
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'request_change_password_otp response without permission');

    // Now assign the "dwarl change own password" permission.
    $auth_role = Role::load(RoleInterface::AUTHENTICATED_ID);
    $this->grantPermissions($auth_role, ['dwarl change own password']);

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'request_change_password_otp response');
  }

  /**
   * Set password using OTP.
   *
   * @param string $password
   *   The password to set.
   * @param string $otp
   *   The otp code.
   * @param string $token
   *   The JWT token.
   */
  protected function setPasswordWithOtpRequest(string $password, string $otp, string $token): void {

    $url = Url::fromRoute('dwarl.set_password_otp');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'password' => $password,
        'otp' => $otp,
      ],
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'request_change_password_otp response');
  }

  /**
   * Set password with passkey.
   *
   * @param string $token
   *   The JWT token.
   * @param string $password
   *   The password.
   * @param string|null $current_password
   *   The current password.
   */
  protected function setPasswordWithPasskey(string $token, string $password, ?string $current_password = NULL): void {
    $url = Url::fromRoute('dwarl.request_set_password_options');
    $request_options = [
      RequestOptions::FORM_PARAMS => ['current_password' => $current_password],
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'request_set_password_options response');

    // Now assign the "dwarl change own password" permission.
    $auth_role = Role::load(RoleInterface::AUTHENTICATED_ID);
    $this->grantPermissions($auth_role, ['dwarl change own password']);

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'request_set_password_options response');

    $data = Json::decode($response->getBody());

    $url = Url::fromRoute('dwarl.set_password_request');
    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'authentication_request' => $data['data']['challenge'],
        'password' => $password,
      ],
      RequestOptions::HEADERS => ['Authorization' => "Bearer $token"],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'set_password_request response');
  }
}
