<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\user\Entity\Role;
use Drupal\user\Entity\User;
use Drupal\user\RoleInterface;
use Drupal\user\UserInterface;
use GuzzleHttp\RequestOptions;

/**
 * Methods for register tests.
 */
trait RegisterTrait {

  /**
   * Create user with email
   *
   * @param string $email
   *   The email.
   * @param bool $setPermission
   *   Should permission be set (and tested)?
   *
   * @return string
   *   The session id.
   */
  protected function createUserEmail(string $email, bool $setPermission): string {
    $url = Url::fromRoute('dwarl.registration.set_email');
    $request_options = [
      RequestOptions::FORM_PARAMS => ['email' => $email],
    ];

    if ($setPermission) {
      // Make request without permission set.
      $response = $this->makeApiRequest('POST', $url, $request_options);

      // Expect access denied.
      $this->assertSame(403, $response->getStatusCode(), 'registration.set_email expect access denied');

      // Assign 'dwarl create account email' permission.
      $anon_role = Role::load(RoleInterface::ANONYMOUS_ID);
      $this->grantPermissions($anon_role, ['dwarl create account email']);
    }

    // Make request again.
    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Check for 200 and token in response.
    $this->assertSame(200, $response->getStatusCode(), 'registration.set_email is 200');

    $headers = $response->getHeaders();
    $this->assertArrayHasKey('Session-Id', $headers);

    // Check user exists.
    $user = user_load_by_mail($email);
    $this->assertInstanceOf(UserInterface::class, $user, 'user exists');

    return $headers['Session-Id'][0];
  }

  /**
   * Create user with phone.
   *
   * @param string $phone
   *   The phone number.
   * @param bool $setPermission
   *   Should permission be set (and tested)?
   *
   * @return string
   *   The session id.
   */
  protected function createUserPhone(string $phone, bool $setPermission): string {
    $url = Url::fromRoute('dwarl_phone.registration.set_phone');
    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'phone' => $phone,
        'country' => 'GB',
      ],
    ];

    if ($setPermission) {
      // Make request without permission set.
      $response = $this->makeApiRequest('POST', $url, $request_options);

      // Expect access denied.
      $this->assertSame(403, $response->getStatusCode(), 'registration.set_phone expect access denied');

      // Assign 'dwarl create account phone' permission.
      $anon_role = Role::load(RoleInterface::ANONYMOUS_ID);
      $this->grantPermissions($anon_role, ['dwarl create account phone']);
      $auth_role = Role::load(RoleInterface::AUTHENTICATED_ID);
      $this->grantPermissions($auth_role, ['dwarl create account phone']);
    }

    // Make request again.
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $data = Json::decode($response->getBody());
    if (isset($data['error'])) {
      print_r($data);
    }

    // Check for 200 and token in response.
    $this->assertSame(200, $response->getStatusCode(), 'registration.set_phone is 200');

    $headers = $response->getHeaders();

    $this->assertArrayHasKey('Session-Id', $headers);


    // Check user exists.
    $user = $this->userService()->loadByUserHandle($phone);
    $this->assertInstanceOf(UserInterface::class, $user, 'user exists');

    return $headers['Session-Id'][0];
  }

  /**
   * Validate contact method.
   *
   * @param string $userHandle
   *   The user handle.
   * @param string $sessionId
   *   The session id.
   * @param string $otpCode
   *   The otp code.
   *
   * @return \Drupal\user\UserInterface
   *   The user.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function validateContactMethod(string $userHandle, string $sessionId, string $otpCode): UserInterface {
    // Test verifying email
    $url = Url::fromRoute('dwarl.registration.verify_contact_method');

    // Try to verify without session id token.
    $request_options = [
      RequestOptions::FORM_PARAMS => ['otp' => $otpCode],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'without session id');

    // Try to verify without otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'without otp');

    // Try to verify with wrong otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['otp' => '4FX2KY'],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'wrong otp');

    // Try to verify with session id and otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['otp' => $otpCode],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Check for 200 and token in response.
    $this->assertSame(200, $response->getStatusCode(), 'session id & otp response code');

    // User account should only be active after auth method has been added.
    $user = $this->userService()->loadByUserHandle($userHandle);
    \Drupal::entityTypeManager()->getStorage('user')->resetCache([$user->id()]);
    $user = User::load($user->id());

    $this->assertTrue($user->isBlocked(), 'user is blocked');

    return $user;
  }

  /**
   * Set password.
   *
   * @param string $password
   *   The password.
   * @param string $sessionId
   *   The session id.
   *
   * @return string
   *   The JWT token.
   */
  protected function registerSetPassword(string $password, string $sessionId): string {
    $url = Url::fromRoute('dwarl.registration.set_password');

    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['password' => $password],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $statusCode = $response->getStatusCode();
    $this->assertSame(200, $statusCode, 'set_password response code');

    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('token', $data, 'set_password has token');
    return $data['token'];
  }

  /**
   * Check passkey info is valid.
   *
   * @param array $passkey
   *   The passkey info.
   */
  protected function assertPasskeyIsValid(array $passkey): void {
    $this->assertArrayHasKey('id', $passkey);
    $this->assertArrayHasKey('created', $passkey);
    $this->assertArrayHasKey('last_used', $passkey);
    $this->assertEquals('dwarl_test os 1.0 dwarl_test client', $passkey['created_device']);
    $this->assertEquals('dwarl_test os 1.0 dwarl_test client', $passkey['last_used_device']);
  }

}
