<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Traits;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\dwarl_test\WebAuthn\RegisterDevice;
use Drupal\dwarl_test\WebAuthn\RequestOptions as DwarlTestRequestOptions;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use GuzzleHttp\RequestOptions;

/**
 * WebAuthn related methods.
 *
 * @todo: Add trying to register as existing user test.
 */
trait WebAuthnTrait {

  /**
   * Register device options.
   *
   * @param string $sessionId
   *   The session id.
   * @param string $route
   *   The route to use.
   *
   * @return array
   *   The device options.
   */
  protected function registerDeviceOptions(string $sessionId, string $route): array {
    $url = Url::fromRoute($route);

    // Make request with wrong session id.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => 'wrong'],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'register_device_options without session id');

    // Make request with session id.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $statusCode = $response->getStatusCode();

    $this->assertSame(200, $statusCode, 'registration.register_device_options response code');
    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('data', $data, 'RDO has data');
    $data = $data['data'];

    $keys = [
      'rp',
      'user',
      'challenge',
      'pubKeyCredParams',
      'authenticatorSelection',
    ];

    foreach ($keys as $key) {
      $this->assertArrayHasKey($key, $data);
    }

    return $data;
  }

  /**
   * Register device.
   *
   * @param string $sessionId
   *   The session id.
   * @param string $userHandle
   *   The user handle.
   * @param array $optionsJson
   *   THe options json.
   * @param string $route
   *   THe route name.
   *
   * @return string
   *    A JWT token.
   *
   * @todo: Remove $optionsJson or use it.
   */
  protected function registerDevice(
    string $sessionId,
    string $userHandle,
    array $optionsJson,
    string $route
  ): string {

    $url = Url::fromRoute($route);

    // Make request without session id.
    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'user_handle' => $userHandle,
        'attestation' => RegisterDevice::TEST_ATTESTATION_PAYLOAD,
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);

    $this->assertSame(403, $response->getStatusCode(), 'register_device without session id');

    // For this test the AuthenticateRequest service has been swapped out
    // for a test one that just checks for a hard coded authentication request.
    $request_options[RequestOptions::HEADERS] = ['session-id' => $sessionId];

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'registration.register_device response code');
    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('token', $data, 'RD has token');
    return $data['token'];
  }

  /**
   * Authenticate request.
   *
   * @param string $userHandle
   *   User handle.
   * @param string $authRequest
   *   Auth request.
   * @param string $sessionId
   *   Session id.
   *
   * @return string
   *   JWT token.
   */
  public function authenticateRequest(
    string $userHandle,
    string $authRequest,
    string $sessionId
  ): string {
    $url = Url::fromRoute('dwarl.authenticate_request');

    $request_options = [
      RequestOptions::FORM_PARAMS => [
        'user_handle' => $userHandle,
        'authentication_request' => $authRequest,
      ],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'authenticate_request without session');

    $request_options[RequestOptions::HEADERS]['session-id'] = $sessionId;
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'authenticate_request');

    $data = Json::decode($response->getBody());

    $this->assertArrayHasKey('token', $data);

    return $data['token'];
  }

  /**
   * Request options.
   *
   * @param string $userHandle
   *   User handle.
   *
   * @return array
   *   Authentication request and session id.
   */
  public function requestOptions(string $userHandle): array {
    $url = Url::fromRoute('dwarl.request_options');

    $request_options = [
      RequestOptions::FORM_PARAMS => ['user_handle' => $userHandle],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'request_options without perm');

    // Assign 'dwarl login with passkey' permission.
    $anon_role = Role::load(RoleInterface::ANONYMOUS_ID);
    $this->grantPermissions($anon_role, ['dwarl login with passkey']);

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'request_options with perm');

    $data = Json::decode($response->getBody());

    $this->assertEquals(
      DwarlTestRequestOptions::TEST_CHALLENGE,
      $data['data']['challenge'],
      'Expect challenge'
    );

    $headers = $response->getHeaders();
    $this->assertArrayHasKey('Session-Id', $headers);

    return [
      'authentication_request' => $data['data']['challenge'],
      'session_id' => reset($headers['Session-Id']),
    ];
  }

}
