<?php

declare(strict_types=1);

namespace Drupal\Tests\dwarl\Functional;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\Tests\ApiRequestTrait;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\dwarl\StartRegistration;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use Drupal\user\UserInterface;
use GuzzleHttp\RequestOptions;

class PasskeyTest extends BrowserTestBase {

  use ApiRequestTrait;

  private const OTP_CODE_LENGTH = 6;

  /**
   * Modules installed for all tests.
   *
   * @var array
   */
  protected static $modules = [
    'system',
    'user',
    'router_test',
    'key',
    'jwt',
    'jwt_test',
    'jwt_auth_issuer',
    'jwt_auth_consumer',
    'public_key_credential_source',
    'webauthn_framework',
    'dwarl',
    'dwarl_test',
  ];

  public function setUp(): void {
    parent::setUp();
  }

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  public function testRegisterAndLogin() {

    // Set test email.
    $email = $this->randomMachineName() . '@email.com';

    $user = user_load_by_mail($email);
    $this->assertFalse($user, "user $email does not exist at start of test");

    // Create user.
    $sessionId = $this->createUserEmail($email);

    // Get users otp code.
    $otp = $this->getOtp($sessionId);

    // Validate email.
    $this->validateUserEmail($email, $sessionId, $otp);

    $optionsJson = $this->registerDeviceOptions($sessionId);
    $jwtToken = $this->registerDevice($sessionId, $email, $optionsJson);

    $this->getPasskeys($jwtToken);
  }

  private function createUserEmail(string $email) {
    $url = Url::fromRoute('dwarl.register.email');
    $request_options = [
      RequestOptions::FORM_PARAMS => ['email' => $email],
    ];

    // Make request without permission set.
    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Expect access denied.
    $this->assertSame(403, $response->getStatusCode());

    // Assign 'dwarl create account email' permission.
    $anon_role = Role::load(RoleInterface::ANONYMOUS_ID);
    $this->grantPermissions($anon_role, ['dwarl create account email']);

    // Make request again.
    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Check for 200 and token in response.
    $this->assertSame(200, $response->getStatusCode());

    $headers = $response->getHeaders();

    $this->assertArrayHasKey('Session-Id', $headers);

    // Check user exists.
    $user = user_load_by_mail($email);
    $this->assertInstanceOf(UserInterface::class, $user, 'user exists');

    return $headers['Session-Id'][0];
  }

  private function getOtp(string $sessionId) {
    /** @var \Drupal\dwarl\Store\AnonymousStoreInterface $store */
    $store = \Drupal::service('dwarl.anonymous_store');

    $store->init($sessionId);

    $otp = $store->getOtp();

    // Check OTP length.
    $this->assertEquals(self::OTP_CODE_LENGTH, strlen($otp));

    return $otp;
  }

  private function validateUserEmail(string $email, string $sessionId, string $otpCode) {
    // Test verifying email
    $url = Url::fromRoute('dwarl.verify');

    // Try to verify without session id token.
    $request_options = [
      RequestOptions::FORM_PARAMS => ['otp' => $otpCode],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'without session id');

    // Try to verify without otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'without otp');

    // Try to verify with wrong otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['otp' => '4FX2KY'],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(403, $response->getStatusCode(), 'wrong otp');

    // Try to verify with session id and otp.
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['otp' => $otpCode],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Check for 200 and token in response.
    $this->assertSame(200, $response->getStatusCode(), 'session id & otp response code');

    drupal_flush_all_caches();

    // User account should only be active after auth method has been added.
    $user = user_load_by_mail($email);

    $this->assertTrue($user->hasRole('validated email'), 'user has validated email role');
    $this->assertTrue($user->isBlocked(), 'user is blocked');
  }

  private function registerDeviceOptions(string $sessionId) {
    $url = Url::fromRoute('dwarl.register_device_options');

    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'register device options response code');
    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('data', $data, 'RDO has data');
    $data = $data['data'];

    $keys = [
      'rp',
      'user',
      'challenge',
      'pubKeyCredParams',
      'authenticatorSelection',
    ];

    foreach ($keys as $key) {
      $this->assertArrayHasKey($key, $data);
    }

    return $data;
  }

  private function registerDevice(string $sessionId, string $email, array $optionsJson): string {
    // Just call a test endpoint as cannot generate attestation with php (yet).
    $url = Url::fromRoute('dwarl_test.register_device_test');

    // Hmm, think this might be impossible to generate with php?
    $attestation = StartRegistration::startRegistration($optionsJson);
    $request_options = [
      RequestOptions::HEADERS => ['session-id' => $sessionId],
      RequestOptions::FORM_PARAMS => ['user_handle' => $email],
    ];
    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'register device options response code');
    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('token', $data, 'RD has token');
    return $data['token'];
  }

  private function getPasskeys(string $token) {
    drupal_flush_all_caches();

    // Just call a test endpoint as cannot generate attestation with php.
    $url = Url::fromRoute('dwarl.passkey.list');

    $request_options = [
      RequestOptions::HEADERS => [
        'Authorization' => "Bearer $token",
        'Access-Control-Allow-Headers' => 'Authorization',
      ],
    ];

    $response = $this->makeApiRequest('POST', $url, $request_options);

    // Expect access denied.
    $this->assertSame(403, $response->getStatusCode());

    // Assign 'dwarl manage own passkeys' permission.
    $auth_role = Role::load(AccountInterface::AUTHENTICATED_ROLE);
    $this->grantPermissions($auth_role, ['dwarl manage own passkeys']);

    $response = $this->makeApiRequest('POST', $url, $request_options);
    $this->assertSame(200, $response->getStatusCode(), 'get passkeys response code');
    $data = Json::decode($response->getBody());
    $this->assertArrayHasKey('passkeys', $data);

    // It will just be an empty array as we didn't really create a passkey.

  }

}
