<?php

namespace Drupal\Tests\dwarl\Functional;

use Drupal\dwarl\Otp\OtpSendMethod;

class RegisterEmailPasskey extends TestBase {

  /**
   * Test: Register with email and passkey.
   */
  public function testRegisterAndLogin() {

    // Set test email.
    $email = $this->randomMachineName() . '@dwarl.com';

    $user = user_load_by_mail($email);
    $this->assertFalse($user, "user $email does not exist at start of test");

    // Create user.
    $sessionId = $this->createUserEmail($email);

    // Get users otp code.
    $otp = $this->getOtp($email, OtpSendMethod::Email);

    // Validate email.
    $this->validateContactMethod($email, $sessionId, $otp);

    // Create passkey.
    $optionsJson = $this->registerDeviceOptions($sessionId, 'dwarl.registration.register_device_options');
    $jwtToken = $this->registerDevice($sessionId, $email, $optionsJson, 'dwarl.registration.register_device');

    // User should have 1 passkey.
    $user = $this->dwarlUser()->loadByUserHandle($email);
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertPasskeyIsValid($passkeys[0]);

    // Fetch user info.
    $userInfo = $this->getUserInfo($jwtToken);
    $this->validateUserInfo($userInfo, $email);

    // Login to fetch new JWT token.
    $response = $this->requestOptions($email);
    $newJwtToken = $this->authenticateRequest($email, $response['authentication_request'], $response['session_id']);

    // Ensure new JTW token works.
    $this->getUserInfo($newJwtToken);

    // Lost passkey flow.
    $sessionId = $this->lostPasskey($email);
    $otp = $this->getOtp($email, OtpSendMethod::Email);
    $status = $this->lostPasskeySubmitOtpOld($otp, $sessionId);

    $this->assertSame('set auth method', $status, 'status should be set auth method as no password is set');

    // Users passkey should have been deleted at this point.
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');

    // We could now set a passkey or a password.
    // Let's create a passkey.
    $optionsJson = $this->registerDeviceOptions($sessionId, 'dwarl.lost_passkey.register_device_options');

    $jwtToken = $this->registerDevice($sessionId, $email, $optionsJson, 'dwarl.lost_passkey.register_device');

    // User should now 1 have passkey again.
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertPasskeyIsValid($passkeys[0]);

    // Set password with passkey.
    $password = 'GNUTerryPratchett';
    $this->setPasswordWithPasskey($jwtToken, $password);

    // Check user has password.
    drupal_flush_all_caches();
    $user = $this->dwarlUser()->loadByUserHandle($email);
    $this->assertSame(60, strlen($user->getPassword()), 'user password should be 60 characters');

    // Do lost passkey flow again, this time with a password set.
    $sessionId = $this->lostPasskey($email);
    $otp = $this->getOtp($email, OtpSendMethod::Email);
    $status = $this->lostPasskeySubmitOtpOld($otp, $sessionId);
    $this->assertSame('set auth method', $status, 'supply password');

    // Users passkey should have been deleted at this point.
    $passkeys = $this->passkeyStorage()->getPasskeys((int) $user->id());
    $this->assertEmpty($passkeys, 'passkeys should be empty');

    $this->lostPasskeySubmitPassword($password, $sessionId);

    // dwarl.lost_passkey.submit_password

    print_r([
      '$status' => $status,
    ]);

  }

  protected function validateUserInfo(array $userInfo, string $email) {
    $this->assertSame($email, $userInfo['email'], 'User info email matches');
    $this->assertNull($userInfo['phone'], 'Phone is not set');
    $this->assertNull($userInfo['preferredContactMethod'], 'preferredContactMethod is not set');
    $this->assertFalse($userInfo['hasPassword'], 'hasPassword is false');
    $this->assertEquals(1, count($userInfo['passkeys']), 'there is one passkey');
    $passkey = reset($userInfo['passkeys']);
    $this->assertPasskeyIsValid($passkey);
  }

}
